from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, cast

from OpenSSL import SSL
from twisted.internet.ssl import (
    AcceptableCiphers,
    CertificateOptions,
    TLSVersion,
    optionsForClientTLS,
)
from twisted.web.client import BrowserLikePolicyForHTTPS
from twisted.web.iweb import IPolicyForHTTPS
from zope.interface.declarations import implementer

from scrapy.core.downloader.tls import (
    _TWISTED_VERSION_MAP,
    _openssl_methods,
    _ScrapyClientTLSOptions,
    _ScrapyClientTLSOptions26,
)
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.utils._deps_compat import TWISTED_TLS_NEW_IMPL
from scrapy.utils.deprecate import create_deprecated_class
from scrapy.utils.misc import build_from_crawler, load_object
from scrapy.utils.ssl import (
    _get_cert_options_version_kwargs,
    _get_keylog_filename,
    _get_tls_version_limits,
    _set_keylog_callback,
)

if TYPE_CHECKING:
    from twisted.internet._sslverify import ClientTLSOptions

    # typing.Self requires Python 3.11
    from typing_extensions import Self

    from scrapy.crawler import Crawler
    from scrapy.settings import BaseSettings


@implementer(IPolicyForHTTPS)
class _ScrapyClientContextFactory(BrowserLikePolicyForHTTPS):
    """Non-peer-certificate verifying HTTPS context factory.

    Uses :setting:`DOWNLOADER_CLIENT_TLS_CIPHERS`,
    :setting:`DOWNLOAD_TLS_MIN_VERSION` and :setting:`DOWNLOAD_TLS_MAX_VERSION`
    to configure the :class:`~twisted.internet.ssl.CertificateOptions`
    instance.

    The purpose of this custom class is to provide a ``creatorForNetloc()``
    method that returns:

    - a ``_ScrapyClientTLSOptions26`` or ``_ScrapyClientTLSOptions`` instance
      configured based on TLS settings provided to the factory (when the
      certificate verification is disabled);
    - a result of ``optionsForClientTLS()`` called with those TLS settings
      (when the certificate verification is enabled).
    """

    def __init__(
        self,
        method: int | None = SSL.SSLv23_METHOD,  # noqa: S503
        tls_verbose_logging: bool = False,
        tls_ciphers: str | None = None,
        *args: Any,
        verify_certificates: bool = False,
        tls_min_version: TLSVersion | None = None,
        tls_max_version: TLSVersion | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)  # type: ignore[no-untyped-call]
        self._ssl_method: int | None = method
        self.tls_min_version: TLSVersion | None = tls_min_version
        self.tls_max_version: TLSVersion | None = tls_max_version
        self.tls_verbose_logging: bool = tls_verbose_logging  # unused
        self.tls_ciphers: AcceptableCiphers | None = (
            AcceptableCiphers.fromOpenSSLCipherString(tls_ciphers)
            if tls_ciphers
            else None
        )
        self._verify_certificates = verify_certificates

    @classmethod
    def from_crawler(
        cls,
        crawler: Crawler,
        method: int | None = SSL.SSLv23_METHOD,  # noqa: S503
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        tls_verbose_logging: bool = crawler.settings.getbool(
            "DOWNLOADER_CLIENT_TLS_VERBOSE_LOGGING"
        )
        tls_ciphers: str | None = crawler.settings["DOWNLOADER_CLIENT_TLS_CIPHERS"]
        # DOWNLOADER_CLIENT_TLS_METHOD reading and handling should be also moved here
        # when the deprecated load_context_factory_from_settings() is removed
        tls_min_ver, tls_max_ver = _get_tls_version_limits(
            crawler.settings, _TWISTED_VERSION_MAP.__getitem__
        )
        if tls_min_ver or tls_max_ver:
            method = None
        verify_certificates = crawler.settings.getbool("DOWNLOAD_VERIFY_CERTIFICATES")
        return cls(  # type: ignore[misc]
            *args,
            method=method,
            tls_verbose_logging=tls_verbose_logging,
            tls_ciphers=tls_ciphers,
            tls_min_version=tls_min_ver,
            tls_max_version=tls_max_ver,
            verify_certificates=verify_certificates,
            **kwargs,
        )

    # should be removed together with ScrapyClientContextFactory
    def getCertificateOptions(self) -> CertificateOptions:  # pragma: no cover
        return self._get_cert_options()

    def _get_cert_options(self) -> CertificateOptions:
        return _ScrapyCertificateOptions(**self._get_cert_options_kwargs())

    def _get_cert_options_kwargs(self) -> dict[str, Any]:
        kwargs: dict[str, Any] = {
            "fixBrokenPeers": True,
            "acceptableCiphers": self.tls_ciphers,
        }
        if self.tls_min_version or self.tls_max_version:
            kwargs.update(
                _get_cert_options_version_kwargs(
                    self.tls_min_version, self.tls_max_version
                )
            )
        # when ScrapyClientContextFactory is removed self._ssl_method can just be None by default
        elif self._ssl_method != SSL.SSLv23_METHOD:
            kwargs["method"] = self._ssl_method
        return kwargs

    # should be removed together with ScrapyClientContextFactory
    def getContext(
        self, hostname: Any = None, port: Any = None
    ) -> SSL.Context:  # pragma: no cover
        return self._get_context()

    def _get_context(self) -> SSL.Context:
        return self._get_cert_options().getContext()

    def creatorForNetloc(self, hostname: bytes, port: int) -> ClientTLSOptions:
        if not self._verify_certificates:
            # Our options class is needed to skip verification errors
            if TWISTED_TLS_NEW_IMPL:
                return _ScrapyClientTLSOptions26(
                    self._get_cert_options()._makeTLSConnection,
                    hostname.decode("ascii"),
                )
            return _ScrapyClientTLSOptions(
                hostname.decode("ascii"),  # type: ignore[arg-type]
                self._get_context(),  # type: ignore[arg-type]
            )
        # Otherwise use the normal Twisted function.
        creator = optionsForClientTLS(
            hostname=hostname.decode("ascii"),
            extraCertificateOptions=self._get_cert_options_kwargs(),
        )
        if _get_keylog_filename():
            # This creator builds its own certificate options, so its context
            # doesn't come from _ScrapyCertificateOptions.
            _set_keylog_callback(_get_creator_context(creator))
        return cast("ClientTLSOptions", creator)


def _get_creator_context(creator: Any) -> SSL.Context:
    """Return the context that *creator* uses for its connections."""
    if TWISTED_TLS_NEW_IMPL:
        return cast("SSL.Context", creator._createConnection.__self__.getContext())
    return cast("SSL.Context", creator._ctx)


ScrapyClientContextFactory = create_deprecated_class(
    "ScrapyClientContextFactory",
    _ScrapyClientContextFactory,
    subclass_warn_message="{old} is deprecated.",
    instance_warn_message="{cls} is deprecated.",
)


@implementer(IPolicyForHTTPS)
class BrowserLikeContextFactory(_ScrapyClientContextFactory):
    """
    Twisted-recommended context factory for web clients.

    Quoting the documentation of the :class:`~twisted.web.client.Agent` class:

        The default is to use a
        :class:`~twisted.web.client.BrowserLikePolicyForHTTPS`, so unless you
        have special requirements you can leave this as-is.

    :meth:`creatorForNetloc` is the same as
    :class:`~twisted.web.client.BrowserLikePolicyForHTTPS` except this context
    factory allows setting the TLS/SSL method to use.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        warnings.warn(
            "BrowserLikeContextFactory is deprecated."
            " You can set DOWNLOAD_VERIFY_CERTIFICATES=True to enable"
            " certificate verification instead of using it.",
            category=ScrapyDeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)

    def creatorForNetloc(self, hostname: bytes, port: int) -> ClientTLSOptions:
        return optionsForClientTLS(  # type: ignore[no-any-return]
            hostname=hostname.decode("ascii"),
            extraCertificateOptions=self._get_cert_options_kwargs(),
        )


@implementer(IPolicyForHTTPS)
class _AcceptableProtocolsContextFactory:
    """Context factory to used to override the acceptable protocols
    to set up the :class:`OpenSSL.SSL.Context` for doing ALPN negotiation.
    It's a private class for :class:`~.H2DownloadHandler`.

    This class wraps ``creatorForNetloc()`` of another factory class, setting
    the acceptable protocols on the :class:`.ClientTLSOptions` instance
    returned by it. It's only needed because we support custom factories via
    :setting:`DOWNLOADER_CLIENTCONTEXTFACTORY`.

    It's a no-op on Twisted 26.4.0+, though using it with custom
    factories on those Twisted versions may be not enough for HTTP/2 support.
    """

    # Something needs to call set_alpn_protos() for ALPN to work.
    #
    # Twisted < 26.4.0 does it in OpenSSLCertificateOptions._makeContext()
    # (requires passing acceptableProtocols from the factory to
    # OpenSSLCertificateOptions) and in TLSMemoryBIOFactory._createConnection()
    # based on H2ClientFactory.acceptableProtocols (too late, it seems).
    #
    # Newer Twisted does it in OpenSSLCertificateOptions._makeContext() as
    # well, and in OpenSSLCertificateOptions._makeTLSConnection() based on
    # H2ClientFactory.acceptableProtocols (which now works).
    #
    # When we drop DOWNLOADER_CLIENTCONTEXTFACTORY it looks like we can replace
    # all of this with _ScrapyClientContextFactory.acceptableProtocols.

    def __init__(self, context_factory: Any, acceptable_protocols: list[bytes]):
        self._wrapped_context_factory: Any = context_factory
        self._acceptable_protocols: list[bytes] = acceptable_protocols

    def creatorForNetloc(self, hostname: bytes, port: int) -> ClientTLSOptions:
        options: ClientTLSOptions = self._wrapped_context_factory.creatorForNetloc(
            hostname, port
        )
        if not TWISTED_TLS_NEW_IMPL:
            from twisted.internet._sslverify import (  # type: ignore[attr-defined]  # noqa: PLC0415  # pylint: disable=no-name-in-module
                _setAcceptableProtocols,
            )

            _setAcceptableProtocols(options._ctx, self._acceptable_protocols)  # type: ignore[attr-defined]
        return options


AcceptableProtocolsContextFactory = create_deprecated_class(
    "AcceptableProtocolsContextFactory",
    _AcceptableProtocolsContextFactory,
    subclass_warn_message="{old} is deprecated.",
    instance_warn_message="{cls} is deprecated.",
)


class _ScrapyCertificateOptions(CertificateOptions):
    """A wrapper needed to add flags to the SSL context before it's used."""

    def _makeContext(self, skipCiphers: bool = False) -> SSL.Context:
        if TWISTED_TLS_NEW_IMPL:
            ctx = super()._makeContext(skipCiphers)
        else:
            ctx = super()._makeContext()
        ctx.set_options(0x4)  # OP_LEGACY_SERVER_CONNECT
        _set_keylog_callback(ctx)
        return ctx


def _load_context_factory_from_settings(crawler: Crawler) -> IPolicyForHTTPS:
    """Create an instance of :setting:`DOWNLOADER_CLIENTCONTEXTFACTORY`.

    Also passes values of other relevant settings to the factory class.
    """
    tls_method_setting: str = crawler.settings["DOWNLOADER_CLIENT_TLS_METHOD"]
    if tls_method_setting != "TLS":
        warnings.warn(
            "Setting DOWNLOADER_CLIENT_TLS_METHOD to a non-default value is"
            " deprecated, please use DOWNLOAD_TLS_MIN_VERSION and/or"
            " DOWNLOAD_TLS_MAX_VERSION instead.",
            ScrapyDeprecationWarning,
            stacklevel=2,
        )
    tls_method = _openssl_methods[tls_method_setting]
    if crawler.settings["DOWNLOADER_CLIENTCONTEXTFACTORY"] == "SENTINEL":
        context_factory_cls = _ScrapyClientContextFactory
    else:  # pragma: no cover
        warnings.warn(
            "The 'DOWNLOADER_CLIENTCONTEXTFACTORY' setting is deprecated.",
            category=ScrapyDeprecationWarning,
            stacklevel=2,
        )
        context_factory_cls = load_object(
            crawler.settings["DOWNLOADER_CLIENTCONTEXTFACTORY"]
        )
    return cast(
        "IPolicyForHTTPS",
        build_from_crawler(
            context_factory_cls,
            crawler,
            method=tls_method,
        ),
    )


def load_context_factory_from_settings(
    settings: BaseSettings, crawler: Crawler
) -> IPolicyForHTTPS:  # pragma: no cover
    warnings.warn(
        "load_context_factory_from_settings() is deprecated.",
        ScrapyDeprecationWarning,
        stacklevel=2,
    )
    return _load_context_factory_from_settings(crawler)



from __future__ import annotations

import socket
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

import pytest
from twisted.conch.telnet import ITelnetProtocol
from twisted.cred import credentials

from scrapy import Spider
from scrapy.extensions.telnet import TelnetConsole, update_telnet_vars
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import Generator

    from scrapy.crawler import Crawler
    from scrapy.http import Response

pytestmark = pytest.mark.requires_reactor  # TelnetConsole requires a reactor


def _get_crawler(
    spidercls: type[Spider] | None = None,
    settings_dict: dict[str, Any] | None = None,
) -> Crawler:
    settings = {
        "TELNETCONSOLE_ENABLED": True,
        **(settings_dict or {}),
    }
    return get_crawler(spidercls, settings_dict=settings)


@contextmanager
def _get_console_and_portal(
    settings: dict[str, Any] | None = None,
) -> Generator[tuple[TelnetConsole, Any]]:
    crawler = _get_crawler(settings_dict=settings)
    console = build_from_crawler(TelnetConsole, crawler)

    # This function has some side effects we don't need for this test
    console._get_telnet_vars = dict  # type: ignore[method-assign]

    console.start_listening()
    protocol = console.protocol()
    portal = protocol.protocolArgs[0]

    try:
        yield console, portal
    finally:
        console.stop_listening()


@coroutine_test
async def test_bad_credentials() -> None:
    with _get_console_and_portal() as (_, portal):
        creds = credentials.UsernamePassword(b"username", b"password")
        d = portal.login(creds, None, ITelnetProtocol)
        with pytest.raises(ValueError, match="Invalid credentials"):
            await maybe_deferred_to_future(d)


@coroutine_test
async def test_good_credentials() -> None:
    with _get_console_and_portal() as (console, portal):
        creds = credentials.UsernamePassword(
            console.username.encode("utf8"), console.password.encode("utf8")
        )
        d = portal.login(creds, None, ITelnetProtocol)
        await maybe_deferred_to_future(d)


@coroutine_test
async def test_custom_credentials() -> None:
    settings = {
        "TELNETCONSOLE_USERNAME": "user",
        "TELNETCONSOLE_PASSWORD": "pass",
    }
    with _get_console_and_portal(settings=settings) as (_, portal):
        creds = credentials.UsernamePassword(b"user", b"pass")
        d = portal.login(creds, None, ITelnetProtocol)
        await maybe_deferred_to_future(d)


def test_invalid_reversed_portrange() -> None:
    settings = {"TELNETCONSOLE_PORT": [2, 1]}
    console = build_from_crawler(TelnetConsole, _get_crawler(settings_dict=settings))
    with pytest.raises(ValueError, match=r"invalid portrange: \[2, 1\]"):
        console.start_listening()


@coroutine_test
async def test_unavailable_port(caplog: pytest.LogCaptureFixture) -> None:
    """Run a crawl where the console cannot bind any port."""
    with socket.create_server(("127.0.0.1", 0)) as sock:
        port = sock.getsockname()[1]
        crawler = _get_crawler(settings_dict={"TELNETCONSOLE_PORT": [port]})
        await crawler.crawl_async()

    assert "CannotListenError" in caplog.text
    assert "AttributeError" not in caplog.text


@coroutine_test
async def test_telnet_vars() -> None:
    """Log into the console of a running crawl, which is when the telnet
    variables are built."""
    received: list[dict[str, Any]] = []

    def on_update_telnet_vars(telnet_vars: dict[str, Any]) -> None:
        received.append(telnet_vars)

    class TelnetSpider(Spider):
        name = "telnet"
        start_urls = ["data:,"]

        async def parse(self, response: Response) -> None:
            console = next(
                ext
                for ext in self.crawler.extensions.middlewares
                if isinstance(ext, TelnetConsole)
            )
            creds = credentials.UsernamePassword(
                console.username.encode("utf8"), console.password.encode("utf8")
            )
            portal = console.protocol().protocolArgs[0]
            await maybe_deferred_to_future(portal.login(creds, None, ITelnetProtocol))

    crawler = _get_crawler(TelnetSpider)
    crawler.signals.connect(on_update_telnet_vars, signal=update_telnet_vars)
    await crawler.crawl_async()

    assert len(received) == 1
    telnet_vars = received[0]
    assert telnet_vars["crawler"] is crawler
    assert telnet_vars["engine"] is crawler.engine
    assert telnet_vars["spider"] is crawler.spider
    assert telnet_vars["extensions"] is crawler.extensions
    assert telnet_vars["stats"] is crawler.stats
    assert telnet_vars["settings"] is crawler.settings
    assert callable(telnet_vars["est"])
    assert callable(telnet_vars["p"])
    assert callable(telnet_vars["prefs"])
    assert "telnetconsole.html" in telnet_vars["help"]



from __future__ import annotations

import re
from typing import TYPE_CHECKING, cast

from scrapy.utils.benchserver import Root, _getarg

if TYPE_CHECKING:
    from twisted.web.server import Request


class _Request:
    def __init__(self, **args: bytes) -> None:
        self.args = {name.encode(): [value] for name, value in args.items()}
        self.written: list[bytes] = []

    def write(self, data: bytes) -> None:
        self.written.append(data)


def test_getarg() -> None:
    request = cast("Request", _Request(total=b"5"))
    assert _getarg(request, b"total", 100, int) == 5
    assert _getarg(request, b"show", 100, int) == 100
    assert _getarg(request, b"missing") is None


def test_render() -> None:
    root = Root()  # type: ignore[no-untyped-call]
    request = _Request(total=b"5", show=b"2")
    assert root.getChild("follow", cast("Request", request)) is root
    assert root.render(cast("Request", request)) == b""
    body = b"".join(request.written).decode()
    assert body.startswith("<html><head></head><body>")
    assert body.endswith("</body></html>")
    numbers = re.findall(
        r"<a href='/follow\?total=5&show=2&n=(\d+)'>follow \1</a>", body
    )
    assert len(numbers) == 2
    assert all(1 <= int(number) <= 5 for number in numbers)



"""Utilities related to asyncio and its support in Scrapy."""

from __future__ import annotations

import asyncio
import logging
import time
from collections.abc import AsyncIterator, Callable, Coroutine, Iterable
from typing import TYPE_CHECKING, Any, Concatenate, ParamSpec, TypeVar

from twisted.internet.defer import Deferred
from twisted.internet.task import LoopingCall, deferLater
from twisted.internet.threads import deferToThread

from scrapy.utils.asyncgen import as_async_generator
from scrapy.utils.reactor import is_asyncio_reactor_installed, is_reactor_installed

if TYPE_CHECKING:
    from twisted.internet.base import DelayedCall

    # typing.Self, typing.TypeVarTuple and typing.Unpack require Python 3.11
    from typing_extensions import Self, TypeVarTuple, Unpack

    _Ts = TypeVarTuple("_Ts")


_T = TypeVar("_T")
_P = ParamSpec("_P")


logger = logging.getLogger(__name__)


def is_asyncio_available() -> bool:
    """Check if it's possible to call asyncio code that relies on the asyncio event loop.

    .. versionadded:: 2.14

    This function returns ``True`` if there is a running asyncio event loop. If
    there is no such loop, it returns ``True`` if the Twisted reactor that is
    installed is
    :class:`~twisted.internet.asyncioreactor.AsyncioSelectorReactor`, returns
    ``False`` if a different reactor is installed, and raises a
    :exc:`RuntimeError` if no reactor is installed.

    Code that doesn't directly require a Twisted reactor should use this
    function while code that requires
    :class:`~twisted.internet.asyncioreactor.AsyncioSelectorReactor` should use
    :func:`~scrapy.utils.reactor.is_asyncio_reactor_installed`.

    When this returns ``True``, an asyncio loop is installed and used by
    Scrapy. It's possible to call functions that require it, such as
    :func:`asyncio.sleep`, and await on :class:`asyncio.Future` objects in
    Scrapy-related code.

    When this returns ``False``, a non-asyncio Twisted reactor is installed.
    It's not possible to use asyncio features that require an asyncio event
    loop or await on :class:`asyncio.Future` objects in Scrapy-related code,
    but it's possible to await on :class:`~twisted.internet.defer.Deferred`
    objects.

    .. note:: As this function uses :func:`asyncio.get_running_loop()`, it will
        only detect the event loop if called in the same thread and from the
        code that runs inside that loop (this shouldn't be a problem when
        calling it from code such as spiders and Scrapy components, if Scrapy
        is run using one of the supported ways).

    .. versionchanged:: 2.15.0
        This function now also returns ``True`` if there is a running asyncio
        loop, even if no Twisted reactor is installed.
    """

    # Check if there is a running asyncio loop.
    # Can't easily check for an installed but not running one, and if we
    # checked that there could be false positives due to some 3rd-party code
    # installing it as a side effect (e.g. by calling get_event_loop()).
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        pass
    else:
        return True

    # Check if there is an installed asyncio reactor (it doesn't need to be
    # running).
    if not is_reactor_installed():
        raise RuntimeError(
            "is_asyncio_available() called without an installed reactor"
            " or running asyncio loop."
        )

    return is_asyncio_reactor_installed()


async def _parallel_asyncio(
    iterable: Iterable[_T] | AsyncIterator[_T],
    count: int,
    callable_: Callable[Concatenate[_T, _P], Coroutine[Any, Any, None]],
    *args: _P.args,
    **kwargs: _P.kwargs,
) -> None:
    """Execute a callable over the objects in the given iterable, in parallel,
    using no more than ``count`` concurrent calls.

    This function is only used in
    :meth:`scrapy.core.scraper.Scraper.handle_spider_output_async` and so it
    assumes that neither *callable* nor iterating *iterable* will raise an
    exception.
    """
    queue: asyncio.Queue[_T | None] = asyncio.Queue(count * 2)

    async def worker() -> None:
        while True:
            item = await queue.get()
            if item is None:
                break
            try:
                await callable_(item, *args, **kwargs)
            finally:
                queue.task_done()

    async def fill_queue() -> None:
        async for item in as_async_generator(iterable):
            await queue.put(item)
        for _ in range(count):
            await queue.put(None)

    fill_task = asyncio.create_task(fill_queue())
    work_tasks = [asyncio.create_task(worker()) for _ in range(count)]
    await asyncio.wait([fill_task, *work_tasks])


class AsyncioLoopingCall:
    """A simple implementation of a periodic call using asyncio, keeping
    some API and behavior compatibility with
    :class:`~twisted.internet.task.LoopingCall`.

    The function is called every *interval* seconds, independent of the finish
    time of the previous call. If the function  is still running when it's time
    to call it again, calls are skipped until the function finishes.

    The function must not return a coroutine or a ``Deferred``.

    .. versionadded:: 2.14.0
    """

    def __init__(self, func: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs):
        self._func: Callable[_P, _T] = func
        self._args: tuple[Any, ...] = args
        self._kwargs: dict[str, Any] = kwargs
        self._task: asyncio.Task[None] | None = None
        self.interval: float | None = None
        self._start_time: float | None = None

    @property
    def running(self) -> bool:
        return self._start_time is not None

    def start(self, interval: float, now: bool = True) -> None:
        """Start calling the function every *interval* seconds.

        :param interval: The interval in seconds between calls.
        :type interval: float

        :param now: If ``True``, also call the function immediately.
        :type now: bool
        """
        if self.running:
            raise RuntimeError("AsyncioLoopingCall already running")

        if interval <= 0:
            raise ValueError("Interval must be greater than 0")

        self.interval = interval
        self._start_time = time.monotonic()
        if now:
            self._call()
        loop = asyncio.get_event_loop()
        self._task = loop.create_task(self._loop())

    def _to_sleep(self) -> float:
        """Return the time to sleep until the next call."""
        assert self.interval is not None
        assert self._start_time is not None
        now = time.monotonic()
        running_for = now - self._start_time
        return self.interval - (running_for % self.interval)

    async def _loop(self) -> None:
        """Run an infinite loop that calls the function periodically."""
        while self.running:
            await asyncio.sleep(self._to_sleep())
            self._call()

    def stop(self) -> None:
        """Stop the periodic calls."""
        self.interval = self._start_time = None
        if self._task is not None:
            self._task.cancel()
            self._task = None

    def _call(self) -> None:
        """Execute the function."""
        try:
            result = self._func(*self._args, **self._kwargs)
        except Exception:
            logger.exception("Error calling the AsyncioLoopingCall function")
            self.stop()
        else:
            if isinstance(result, (Coroutine, Deferred)):
                self.stop()
                raise TypeError(
                    "The AsyncioLoopingCall function must not return a coroutine or a Deferred"
                )


def create_looping_call(
    func: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs
) -> AsyncioLoopingCall | LoopingCall:
    """Create an instance of a looping call class.

    This creates an instance of
    :class:`~scrapy.utils.asyncio.AsyncioLoopingCall` or
    :class:`~twisted.internet.task.LoopingCall`, depending on whether asyncio
    support is available.

    .. versionadded:: 2.14.0
    """
    if is_asyncio_available():
        return AsyncioLoopingCall(func, *args, **kwargs)
    return LoopingCall(func, *args, **kwargs)


def call_later(
    delay: float, func: Callable[[Unpack[_Ts]], object], *args: Unpack[_Ts]
) -> CallLaterResult:
    """Schedule a function to be called after a delay.

    This uses either :meth:`asyncio.loop.call_later` or
    :meth:`reactor.callLater() <twisted.internet.base.ReactorBase.callLater>`,
    depending on whether asyncio support is available.

    .. versionadded:: 2.14.0
    """
    if is_asyncio_available():
        loop = asyncio.get_event_loop()
        return CallLaterResult.from_asyncio(loop.call_later(delay, func, *args))

    from twisted.internet import reactor

    return CallLaterResult.from_twisted(reactor.callLater(delay, func, *args))


class CallLaterResult:
    """An universal result for :func:`call_later`, wrapping either
    :class:`asyncio.TimerHandle` or :class:`twisted.internet.base.DelayedCall`.

    The provided API is close to the :class:`asyncio.TimerHandle` one: there is
    no ``active()`` (as there is no such public API in
    :class:`asyncio.TimerHandle`) but ``cancel()`` can be called on already
    called or cancelled instances.

    .. versionadded:: 2.14.0
    """

    _timer_handle: asyncio.TimerHandle | None = None
    _delayed_call: DelayedCall | None = None

    @classmethod
    def from_asyncio(cls, timer_handle: asyncio.TimerHandle) -> Self:
        """Create a CallLaterResult from an asyncio TimerHandle."""
        o = cls()
        o._timer_handle = timer_handle
        return o

    @classmethod
    def from_twisted(cls, delayed_call: DelayedCall) -> Self:
        """Create a CallLaterResult from a Twisted DelayedCall."""
        o = cls()
        o._delayed_call = delayed_call
        return o

    def cancel(self) -> None:
        """Cancel the underlying delayed call.

        Does nothing if the delayed call was already called or cancelled.
        """
        if self._timer_handle:
            self._timer_handle.cancel()
            self._timer_handle = None
        elif self._delayed_call and self._delayed_call.active():
            self._delayed_call.cancel()
            self._delayed_call = None


async def sleep(seconds: float) -> None:
    """Sleep for *seconds*.

    .. versionadded:: 2.18.0

    This uses either :func:`asyncio.sleep` or
    :func:`~twisted.internet.task.deferLater`, depending on whether asyncio
    support is available.
    """
    if is_asyncio_available():
        await asyncio.sleep(seconds)
        return

    from twisted.internet import reactor

    await deferLater(reactor, seconds)


async def run_in_thread(
    func: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs
) -> _T:
    """Call a function in a thread and return its result as a coroutine.

    This uses either :func:`asyncio.to_thread` or
    :func:`twisted.internet.threads.deferToThread`, depending on whether
    asyncio support is available.

    .. versionadded:: 2.15.0
    """
    if is_asyncio_available():
        return await asyncio.to_thread(func, *args, **kwargs)

    # circular import
    from scrapy.utils.defer import maybe_deferred_to_future  # noqa: PLC0415

    return await maybe_deferred_to_future(deferToThread(func, *args, **kwargs))



"""Helper functions for scrapy.http objects (Request, Response)"""

from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import ParseResult, urlparse
from weakref import WeakKeyDictionary

if TYPE_CHECKING:
    from scrapy.http import Request, Response


_urlparse_cache: WeakKeyDictionary[Request | Response, ParseResult] = (
    WeakKeyDictionary()
)


def urlparse_cached(request_or_response: Request | Response) -> ParseResult:
    """Return the result of parsing the URL of *request_or_response*, a
    :class:`~scrapy.Request` or :class:`~scrapy.http.Response` object, with
    :func:`urllib.parse.urlparse`.

    The result is cached, using a :class:`weakref.WeakKeyDictionary` keyed on
    *request_or_response*, so that the URL of a given object is parsed only
    once. Prefer this function over calling :func:`urllib.parse.urlparse` on
    ``request_or_response.url`` directly when the same URL may be parsed more
    than once.
    """
    if request_or_response not in _urlparse_cache:
        _urlparse_cache[request_or_response] = urlparse(request_or_response.url)
    return _urlparse_cache[request_or_response]



def response(flow) -> None:
    # add custom headers to be able to check that the request went through the proxy
    flow.response.headers["X-Via-Mitmproxy"] = "1"
    if flow.client_conn.tls_established:
        flow.response.headers["X-Via-Mitmproxy-TLS"] = "1"



"""Scrapy Shell

See documentation in docs/topics/shell.rst

"""

from __future__ import annotations

import asyncio
import contextlib
import os
import signal
import warnings
from typing import TYPE_CHECKING, Any

from itemadapter import is_item
from twisted.internet import threads
from twisted.internet.defer import Deferred
from twisted.python import threadable
from w3lib.url import any_to_uri

import scrapy
from scrapy.crawler import Crawler
from scrapy.exceptions import IgnoreRequest, ScrapyDeprecationWarning
from scrapy.http import Request, Response
from scrapy.settings import Settings
from scrapy.spiders import Spider
from scrapy.utils.conf import get_config
from scrapy.utils.console import DEFAULT_PYTHON_SHELLS, start_python_console
from scrapy.utils.datatypes import SequenceExclude
from scrapy.utils.defer import deferred_f_from_coro_f, maybe_deferred_to_future
from scrapy.utils.misc import load_object
from scrapy.utils.response import open_in_browser

if TYPE_CHECKING:
    from collections.abc import Callable

# Hopefully temporary architecture notes
#
# The Shell class is always instantiated in the "main" thread. There are two
# official ways to use it:
# 1. scrapy.commands.shell, which makes a secondary thread and calls
# CrawlerProcess.start() in it, which runs a reactor there.
# 2. scrapy.shell.inspect_response(), which just creates Shell() in the current
# thread.
#
# Shell._inthread is True when this class is run in a thread separate from the
# reactor, e.g. the 1st way (in other words, the reactor is in a secondary
# thread).
# Shell._inthread is False when this class is run in the same thread as the
# reactor, e.g. the 2nd way.
# The only thing that differs is availability of fetch() (it needs the
# reactor to be in a separate thread: the shell sends the request to
# the reactor and waits for the result synchronously).
#
# Thus the only thing Shell needs an event loop for is fetch(). More machinery
# is used for it to work. In chronological order:
# 1. scrapy.commands.shell.Command.run() creates a crawler and an engine, then
# calls
# _schedule_coro(crawler.engine.start_async(_start_request_processing=False)),
# which initializes the engine but doesn't start processing of requests.
# 2. scrapy.commands.shell.Command.run() calls crawler_process.start() in a
# thread which starts a reactor in that thread.
# 3. When fetch() is called, it prepares a request and calls Shell._schedule()
# in the reactor thread (via threads.blockingCallFromThread()).
# 4. Shell._schedule() calls Shell._open_spider() (on the first call).
# 5. Shell._open_spider() calls
# engine.open_spider_async(close_if_idle=False).
# 6. Shell._schedule() calls engine.crawl(request), scheduling the request.
# 7. Shell._schedule() via _request_deferred() waits until the request callback
# is called. When it's called, the response becomes available.
#
# In the reactorless mode this is slightly different, the engine initialization
# happens in the event loop thread as many things need either a reactor or a
# running event loop.
#
# Side note: it should be possible to remove _request_deferred() by using
# engine.download_async() instead of engine.crawl(), losing the usual stuff
# like spider middlewares (none of which should be important).
#
# Other architecture problems:
# * scrapy.cmdline.execute() creates an AsyncCrawlerProcess instance which
#   immediately installs a reactor (which is maybe not thread-specific?) or an
#   event loop (which *is* thread-specific, so the main thread will always have
#   a (not running) loop installed.
# * scrapy.commands.shell.Command.run() calls _schedule_coro() in the main
#   thread, and various engine init code also calls similar things,
#   conceptually this shouldn't work (and doesn't in the reactorless mode, so
#   there the initialization is moved to the event loop thread).
# * The engine has several code paths specifically for the shell, and the shell
#   uses several private members of the engine and of AsyncCrawlerProcess.


class Shell:
    relevant_classes: tuple[type, ...] = (Crawler, Spider, Request, Response, Settings)

    def __init__(
        self,
        crawler: Crawler,
        update_vars: Callable[[dict[str, Any]], None] | None = None,
        code: str | None = None,
        *,
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        self._use_reactor = crawler.settings.getbool("TWISTED_REACTOR_ENABLED")
        if not self._use_reactor and not loop:  # pragma: no cover
            raise RuntimeError(
                "Shell needs the crawler loop reference when TWISTED_REACTOR_ENABLED=False."
            )
        self._loop = loop
        self.crawler: Crawler = crawler
        self.update_vars: Callable[[dict[str, Any]], None] = update_vars or (
            lambda x: None
        )
        self.item_class: type = load_object(crawler.settings["DEFAULT_ITEM_CLASS"])
        self.spider: Spider | None = None
        if self._use_reactor:
            self._inthread: bool = not threadable.isInIOThread()
        else:
            try:
                # in case there is also a running loop in the main thread
                current_loop = asyncio.get_running_loop()
                self._inthread = current_loop is not self._loop
            except RuntimeError:
                self._inthread = True
        self.code: str | None = code
        self.vars: dict[str, Any] = {}

    @property
    def inthread(self) -> bool:  # pragma: no cover
        warnings.warn(
            "Shell.inthread is deprecated, use Shell.fetch_available instead.",
            ScrapyDeprecationWarning,
            stacklevel=2,
        )
        return self._inthread

    @property
    def fetch_available(self) -> bool:
        """Whether fetch() can be used."""
        return self._inthread

    def start(
        self,
        url: str | None = None,
        request: Request | None = None,
        response: Response | None = None,
        spider: Spider | None = None,
        redirect: bool = True,
    ) -> None:
        # disable accidental Ctrl-C key press from shutting down the engine
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        if url:
            self.fetch(url, spider, redirect=redirect)
        elif request:
            self.fetch(request, spider)
        elif response:
            request = response.request
            self.populate_vars(response, request, spider)
        else:
            self.populate_vars()
        if self.code:
            print(eval(self.code, globals(), self.vars))  # noqa: S307
        else:
            # Detect interactive shell setting in scrapy.cfg
            # e.g.: ~/.config/scrapy.cfg or ~/.scrapy.cfg
            # [settings]
            # # shell can be one of ipython, bpython or python;
            # # to be used as the interactive python console, if available.
            # # (default is ipython, fallbacks in the order listed above)
            # shell = python
            cfg = get_config()
            section, option = "settings", "shell"
            env = os.environ.get("SCRAPY_PYTHON_SHELL")
            shells = []
            if env:
                shells += env.strip().lower().split(",")
            elif cfg.has_option(section, option):
                shells += [cfg.get(section, option).strip().lower()]
            else:  # try all by default
                shells += DEFAULT_PYTHON_SHELLS.keys()
            # always add standard shell as fallback
            shells += ["python"]
            start_python_console(
                self.vars, shells=shells, banner=self.vars.pop("banner", "")
            )

    async def _schedule(self, request: Request, spider: Spider | None) -> Response:
        """Send the request to the engine, wait for the result.

        Runs in the reactor thread when using the reactor, or in the asyncio
        event loop thread otherwise.
        """
        if not self.spider:
            await self._open_spider(spider)
        # send the request to the engine
        self.crawler.engine.crawl(request)
        # this will fire when the request callback runs (via the callback hijacking in _request_deferred())
        return await maybe_deferred_to_future(_request_deferred(request))

    async def _open_spider(self, spider: Spider | None) -> None:
        if spider is None:
            spider = self.crawler.spider or self.crawler._create_spider()

        self.crawler.spider = spider
        await self.crawler.engine.open_spider_async(close_if_idle=False)
        self.spider = spider

    def fetch(
        self,
        request_or_url: Request | str,
        spider: Spider | None = None,
        redirect: bool = True,
        **kwargs: Any,
    ) -> None:
        if isinstance(request_or_url, Request):
            request = request_or_url
        else:
            url = any_to_uri(request_or_url)
            request = Request(url, dont_filter=True, **kwargs)
            if redirect:
                request.meta["handle_httpstatus_list"] = SequenceExclude(
                    range(300, 400)
                )
            else:
                request.meta["handle_httpstatus_all"] = True
        response: Response | None = None
        if self._use_reactor:
            from twisted.internet import reactor

            with contextlib.suppress(IgnoreRequest):
                response = threads.blockingCallFromThread(
                    reactor,
                    deferred_f_from_coro_f(self._schedule),  # type: ignore[arg-type]
                    request,
                    spider,
                )
        else:
            assert self._loop
            with contextlib.suppress(IgnoreRequest):
                future = asyncio.run_coroutine_threadsafe(
                    self._schedule(request, spider), self._loop
                )
                response = future.result()
        self.populate_vars(response, request, self.spider)

    def populate_vars(
        self,
        response: Response | None = None,
        request: Request | None = None,
        spider: Spider | None = None,
    ) -> None:
        self.vars["scrapy"] = scrapy
        self.vars["crawler"] = self.crawler
        self.vars["item"] = self.item_class()
        self.vars["settings"] = self.crawler.settings
        self.vars["spider"] = spider
        self.vars["request"] = request
        self.vars["response"] = response
        if self.fetch_available:
            self.vars["fetch"] = self.fetch
        self.vars["view"] = open_in_browser
        self.vars["shelp"] = self.print_help
        self.update_vars(self.vars)
        if not self.code:
            self.vars["banner"] = self.get_help()

    def print_help(self) -> None:
        print(self.get_help())

    def get_help(self) -> str:
        b = []
        b.append("Available Scrapy objects:")
        b.append(
            "  scrapy     scrapy module (contains scrapy.Request, scrapy.Selector, etc)"
        )
        for k, v in sorted(self.vars.items()):
            if self._is_relevant(v):
                b.append(f"  {k:<10} {v}")
        b.append("Useful shortcuts:")
        if self.fetch_available:
            b.append(
                "  fetch(url[, redirect=True]) "
                "Fetch URL and update local objects (by default, redirects are followed)"
            )
            b.append(
                "  fetch(req)                  "
                "Fetch a scrapy.Request and update local objects "
            )
        b.append("  shelp()           Shell help (print this help)")
        b.append("  view(response)    View response in a browser")

        return "\n".join(f"[s] {line}" for line in b) + "\n"

    def _is_relevant(self, value: Any) -> bool:
        return isinstance(value, self.relevant_classes) or is_item(value)


def inspect_response(response: Response, spider: Spider) -> None:
    """Open a shell to inspect the given response"""
    # Shell.start removes the SIGINT handler, so save it and re-add it after
    # the shell has closed
    sigint_handler = signal.getsignal(signal.SIGINT)
    if not spider.crawler.settings.getbool("TWISTED_REACTOR_ENABLED"):
        loop = asyncio.get_running_loop()
    else:
        loop = None
    Shell(spider.crawler, loop=loop).start(response=response, spider=spider)
    signal.signal(signal.SIGINT, sigint_handler)


def _request_deferred(request: Request) -> Deferred[Any]:
    """Wrap a request inside a Deferred.

    This function is harmful, do not use it until you know what you are doing.

    This returns a Deferred whose first pair of callbacks are the request
    callback and errback. The Deferred also triggers when the request
    callback/errback is executed (i.e. when the request is downloaded)

    WARNING: Do not call request.replace() until after the deferred is called.
    """
    request_callback = request.callback
    request_errback = request.errback

    def _restore_callbacks(result: Any) -> Any:
        request.callback = request_callback
        request.errback = request_errback
        return result

    d: Deferred[Any] = Deferred()
    d.addBoth(_restore_callbacks)
    if request.callback:
        d.addCallback(request.callback)
    if request.errback:
        d.addErrback(request.errback)

    request.callback, request.errback = d.callback, d.errback
    return d



"""This module provides some functions and classes to record and report
references to live object instances.

If you want live objects for a particular class to be tracked, you only have to
subclass from object_ref (instead of object).

This library has a minimal performance impact.

.. note:: PyPy uses a tracing garbage collector, so objects may
    remain in the ``live_refs`` longer than expected, even after they
    go out of scope. If deterministic behavior is required, you may need
    to explicitly trigger garbage collection or call ``trackref.live_refs.clear()``.
"""

from __future__ import annotations

from operator import itemgetter
from time import monotonic_ns
from types import NoneType
from typing import TYPE_CHECKING, Any
from weakref import WeakKeyDictionary

if TYPE_CHECKING:
    from collections.abc import Iterable

    # typing.Self requires Python 3.11
    from typing_extensions import Self


live_refs: WeakKeyDictionary[type, WeakKeyDictionary[object, float]] = (
    WeakKeyDictionary()
)


class object_ref:
    """Inherit from this class if you want to track live instances with the
    ``trackref`` module."""

    __slots__ = ()

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        obj = object.__new__(cls)
        try:
            refs = live_refs[cls]
        except KeyError:
            refs = live_refs[cls] = WeakKeyDictionary()
        refs[obj] = monotonic_ns()
        return obj


# using Any as it's hard to type type(None)
def format_live_refs(ignore: Any = NoneType) -> str:
    """Return a tabular representation of tracked objects"""
    s = "Live References\n\n"
    now_ns = monotonic_ns()
    for cls, wdict in sorted(live_refs.items(), key=lambda x: x[0].__name__):
        if not wdict:
            continue
        if issubclass(cls, ignore):
            continue
        oldest_ns = min(wdict.values())
        s += f"{cls.__name__:<30} {len(wdict):6}   oldest: {int((now_ns - oldest_ns) // 1e9)}s ago\n"
    return s


def print_live_refs(*a: Any, **kw: Any) -> None:
    """Print a report of live references, grouped by class name.

    :param ignore: if given, all objects from the specified class (or tuple of
        classes) will be ignored.
    :type ignore: type or tuple
    """
    print(format_live_refs(*a, **kw))


def get_oldest(class_name: str) -> Any:
    """Return the oldest object alive with the given class name, or ``None`` if
    none is found. Use :func:`print_live_refs` first to get a list of all
    tracked live objects per class name."""
    for cls, wdict in live_refs.items():
        if cls.__name__ == class_name:
            if not wdict:
                break
            return min(wdict.items(), key=itemgetter(1))[0]
    return None


def iter_all(class_name: str) -> Iterable[Any]:
    """Return an iterator over all objects alive with the given class name. Use
    :func:`print_live_refs` first to get a list of all tracked live objects per
    class name."""
    for cls, wdict in live_refs.items():
        if cls.__name__ == class_name:
            return wdict.keys()
    return ()



from __future__ import annotations

import gc

import pytest

from scrapy.exceptions import NotConfigured
from scrapy.extensions.memdebug import MemoryDebugger
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.spider import DefaultSpider
from scrapy.utils.test import get_crawler
from scrapy.utils.trackref import object_ref
from tests.utils.decorators import coroutine_test


def test_disabled_by_default() -> None:
    with pytest.raises(NotConfigured):
        build_from_crawler(MemoryDebugger, get_crawler())


def test_spider_closed_sets_stats() -> None:
    crawler = get_crawler(settings_dict={"MEMDEBUG_ENABLED": True})
    ext = build_from_crawler(MemoryDebugger, crawler)

    class TrackedObject(object_ref):
        pass

    class CollectedObject(object_ref):
        pass

    tracked = [TrackedObject(), TrackedObject()]
    CollectedObject()

    ext.spider_closed(DefaultSpider.from_crawler(crawler), "finished")

    assert crawler.stats.get_value("memdebug/gc_garbage_count") == len(gc.garbage)
    assert crawler.stats.get_value("memdebug/live_refs/TrackedObject") == len(tracked)
    assert crawler.stats.get_value("memdebug/live_refs/CollectedObject") is None


@coroutine_test
async def test_crawl_sets_stats() -> None:
    # unique class so that other tests don't pollute live_refs
    class MemDebugSpider(DefaultSpider):
        pass

    crawler = get_crawler(MemDebugSpider, settings_dict={"MEMDEBUG_ENABLED": True})
    await crawler.crawl_async()
    assert crawler.stats.get_value("memdebug/gc_garbage_count") is not None
    assert crawler.stats.get_value("memdebug/live_refs/MemDebugSpider") == 1



from __future__ import annotations

import asyncio
import logging
import re
import signal
import threading
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar
from unittest.mock import MagicMock

import pytest

import scrapy
from scrapy import Spider
from scrapy.crawler import AsyncCrawlerProcess, Crawler, CrawlerProcess
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.extensions.throttle import AutoThrottle
from scrapy.settings import Settings, default_settings
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.log import (
    _uninstall_scrapy_root_handler,
    configure_logging,
    get_scrapy_root_handler,
)
from scrapy.utils.spider import DefaultSpider
from scrapy.utils.test import get_crawler, get_reactor_settings
from tests.spiders import NoRequestsSpider
from tests.utils import assert_option_is_default
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import Callable

BASE_SETTINGS: dict[str, Any] = {}


def get_raw_crawler(
    spidercls: type[Spider] | None = None, settings_dict: dict[str, Any] | None = None
) -> Crawler:
    """get_crawler alternative that only calls the __init__ method of the
    crawler."""
    settings = Settings()
    settings.setdict(get_reactor_settings())
    settings.setdict(settings_dict or {})
    return Crawler(spidercls or DefaultSpider, settings)


class TestCrawler:
    def test_populate_spidercls_settings(self) -> None:
        spider_settings: dict[str, Any] = {
            "TEST1": "spider",
            "TEST2": "spider",
        }
        project_settings = {
            **BASE_SETTINGS,
            "TEST1": "project",
            "TEST3": "project",
            **get_reactor_settings(),
        }

        class CustomSettingsSpider(DefaultSpider):
            custom_settings = spider_settings

        settings = Settings()
        settings.setdict(project_settings, priority="project")
        crawler = Crawler(CustomSettingsSpider, settings)
        crawler._apply_settings()

        assert crawler.settings.get("TEST1") == "spider"
        assert crawler.settings.get("TEST2") == "spider"
        assert crawler.settings.get("TEST3") == "project"

        assert not settings.frozen
        assert crawler.settings.frozen

    @pytest.mark.parametrize(
        "attr",
        ["extensions", "logformatter", "request_fingerprinter", "stats"],
    )
    def test_late_attr_before_apply_settings(self, attr: str) -> None:
        crawler = get_raw_crawler(DefaultSpider)
        with pytest.raises(RuntimeError, match=rf"Crawler\.{attr} is not set yet"):
            getattr(crawler, attr)
        crawler._apply_settings()
        assert getattr(crawler, attr) is not None

    @pytest.mark.parametrize(
        "attr",
        ["engine", "extensions", "logformatter", "request_fingerprinter", "stats"],
    )
    def test_late_attr_on_class(self, attr: str) -> None:
        # Introspection tools such as help() read these off the class.
        assert getattr(Crawler, attr) is getattr(Crawler, attr)

    def test_late_attr_engine_before_crawl(self) -> None:
        crawler = get_raw_crawler(DefaultSpider)
        crawler._apply_settings()
        with pytest.raises(RuntimeError, match=r"Crawler\.engine is not set yet"):
            _ = crawler.engine

    @pytest.mark.parametrize(
        ("attr", "setting"),
        [
            ("download_delay", "DOWNLOAD_DELAY"),
            ("max_concurrent_requests", "CONCURRENT_REQUESTS_PER_DOMAIN"),
        ],
    )
    def test_deprecated_spider_attr(self, attr: str, setting: str) -> None:
        crawler = get_raw_crawler(type("_Spider", (DefaultSpider,), {attr: 2}))
        with pytest.warns(
            ScrapyDeprecationWarning,
            match=f"The {attr!r} spider attribute is deprecated. Use the {setting} ",
        ):
            crawler._apply_settings()
        assert crawler.settings.getint(setting) == 2

    @pytest.mark.parametrize(
        ("attr", "setting"),
        [
            ("download_delay", "DOWNLOAD_DELAY"),
            ("max_concurrent_requests", "CONCURRENT_REQUESTS_PER_DOMAIN"),
        ],
    )
    def test_deprecated_spider_attr_ignored(self, attr: str, setting: str) -> None:
        crawler = get_raw_crawler(type("_Spider", (DefaultSpider,), {attr: 2}))
        crawler.settings.set(setting, 3, priority="spider")
        with pytest.warns(
            ScrapyDeprecationWarning,
            match=f"The {attr!r} spider attribute is deprecated. It is also being ",
        ):
            crawler._apply_settings()
        assert crawler.settings.getint(setting) == 3

    def test_crawler_accepts_dict(self) -> None:
        crawler = get_crawler(DefaultSpider, {"foo": "bar"})
        assert crawler.settings["foo"] == "bar"
        assert_option_is_default(crawler.settings, "RETRY_ENABLED")

    def test_crawler_accepts_None(self) -> None:
        crawler = Crawler(DefaultSpider)
        assert_option_is_default(crawler.settings, "RETRY_ENABLED")

    def test_crawler_rejects_spider_objects(self) -> None:
        with pytest.raises(ValueError, match="spidercls argument must be a class"):
            Crawler(DefaultSpider())  # type: ignore[arg-type]

    @coroutine_test
    async def test_crawler_crawl_twice_seq_unsupported(self) -> None:
        crawler = get_raw_crawler(NoRequestsSpider, BASE_SETTINGS)
        await maybe_deferred_to_future(crawler.crawl())
        with pytest.raises(RuntimeError, match="more than once on the same instance"):
            await maybe_deferred_to_future(crawler.crawl())

    @coroutine_test
    async def test_crawler_crawl_async_twice_seq_unsupported(self) -> None:
        crawler = get_raw_crawler(NoRequestsSpider, BASE_SETTINGS)
        await crawler.crawl_async()
        with pytest.raises(RuntimeError, match="more than once on the same instance"):
            await crawler.crawl_async()

    @coroutine_test
    async def test_crawler_crawl_twice_parallel_unsupported(self) -> None:
        crawler = get_raw_crawler(NoRequestsSpider, BASE_SETTINGS)
        d1 = crawler.crawl()
        d2 = crawler.crawl()
        await maybe_deferred_to_future(d1)
        with pytest.raises(RuntimeError, match="Crawling already taking place"):
            await maybe_deferred_to_future(d2)

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_crawler_crawl_async_twice_parallel_unsupported(self) -> None:
        crawler = get_raw_crawler(NoRequestsSpider, BASE_SETTINGS)
        t1 = asyncio.create_task(crawler.crawl_async())
        t2 = asyncio.create_task(crawler.crawl_async())
        await t1
        with pytest.raises(RuntimeError, match="Crawling already taking place"):
            await t2

    def test_get_addon(self) -> None:
        class ParentAddon:
            pass

        class TrackingAddon(ParentAddon):
            instances: ClassVar[list[TrackingAddon]] = []

            def __init__(self):
                TrackingAddon.instances.append(self)

            def update_settings(self, settings):
                pass

        settings = {
            **BASE_SETTINGS,
            "ADDONS": {
                TrackingAddon: 0,
            },
        }
        crawler = get_crawler(settings_dict=settings)
        assert len(TrackingAddon.instances) == 1
        expected = TrackingAddon.instances[-1]

        addon = crawler.get_addon(TrackingAddon)
        assert addon == expected

        addon = crawler.get_addon(DefaultSpider)  # type: ignore[assignment]
        assert addon is None

        addon = crawler.get_addon(ParentAddon)
        assert addon == expected

        class ChildAddon(TrackingAddon):
            pass

        addon = crawler.get_addon(ChildAddon)
        assert addon is None

    @coroutine_test
    async def test_get_downloader_middleware(self) -> None:
        class ParentDownloaderMiddleware:
            pass

        class TrackingDownloaderMiddleware(ParentDownloaderMiddleware):
            instances: ClassVar[list[TrackingDownloaderMiddleware]] = []

            def __init__(self):
                TrackingDownloaderMiddleware.instances.append(self)

        class MySpider(Spider):
            name = "myspider"
            cls: ClassVar[type[Any]]
            result: ClassVar[Any]

            @classmethod
            def from_crawler(cls, crawler):
                return cls(crawler=crawler)

            def __init__(self, crawler, **kwargs: Any):
                super().__init__(**kwargs)
                self.crawler = crawler

            async def start(self):
                MySpider.result = crawler.get_downloader_middleware(MySpider.cls)
                return
                yield

        settings = {
            **BASE_SETTINGS,
            "DOWNLOADER_MIDDLEWARES": {
                TrackingDownloaderMiddleware: 0,
            },
        }

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = TrackingDownloaderMiddleware
        await crawler.crawl_async()
        assert len(TrackingDownloaderMiddleware.instances) == 1
        assert MySpider.result == TrackingDownloaderMiddleware.instances[-1]

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = DefaultSpider
        await crawler.crawl_async()
        assert MySpider.result is None

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ParentDownloaderMiddleware
        await crawler.crawl_async()
        assert MySpider.result == TrackingDownloaderMiddleware.instances[-1]

        class ChildDownloaderMiddleware(TrackingDownloaderMiddleware):
            pass

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ChildDownloaderMiddleware
        await crawler.crawl_async()
        assert MySpider.result is None

    def test_get_downloader_middleware_not_crawling(self) -> None:
        crawler = get_raw_crawler(settings_dict=BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            crawler.get_downloader_middleware(DefaultSpider)

    @coroutine_test
    async def test_get_downloader_middleware_no_engine(self) -> None:
        class MySpider(Spider):
            name = "myspider"

            @classmethod
            def from_crawler(cls, crawler):
                crawler.get_downloader_middleware(DefaultSpider)

        crawler = get_raw_crawler(MySpider, BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            await crawler.crawl_async()

    @coroutine_test
    async def test_get_extension(self) -> None:
        class ParentExtension:
            pass

        class TrackingExtension(ParentExtension):
            instances: ClassVar[list[TrackingExtension]] = []

            def __init__(self):
                TrackingExtension.instances.append(self)

        class MySpider(Spider):
            name = "myspider"
            cls: ClassVar[type[Any]]
            result: ClassVar[Any]

            @classmethod
            def from_crawler(cls, crawler):
                return cls(crawler=crawler)

            def __init__(self, crawler, **kwargs: Any):
                super().__init__(**kwargs)
                self.crawler = crawler

            async def start(self):
                MySpider.result = crawler.get_extension(MySpider.cls)
                return
                yield

        settings = {
            **BASE_SETTINGS,
            "EXTENSIONS": {
                TrackingExtension: 0,
            },
        }

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = TrackingExtension
        await crawler.crawl_async()
        assert len(TrackingExtension.instances) == 1
        assert MySpider.result == TrackingExtension.instances[-1]

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = DefaultSpider
        await crawler.crawl_async()
        assert MySpider.result is None

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ParentExtension
        await crawler.crawl_async()
        assert MySpider.result == TrackingExtension.instances[-1]

        class ChildExtension(TrackingExtension):
            pass

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ChildExtension
        await crawler.crawl_async()
        assert MySpider.result is None

    def test_get_extension_not_crawling(self) -> None:
        crawler = get_raw_crawler(settings_dict=BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            crawler.get_extension(DefaultSpider)

    @coroutine_test
    async def test_get_extension_no_engine(self) -> None:
        class MySpider(Spider):
            name = "myspider"

            @classmethod
            def from_crawler(cls, crawler):
                crawler.get_extension(DefaultSpider)

        crawler = get_raw_crawler(MySpider, BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            await crawler.crawl_async()

    @coroutine_test
    async def test_get_item_pipeline(self) -> None:
        class ParentItemPipeline:
            pass

        class TrackingItemPipeline(ParentItemPipeline):
            instances: ClassVar[list[TrackingItemPipeline]] = []

            def __init__(self):
                TrackingItemPipeline.instances.append(self)

        class MySpider(Spider):
            name = "myspider"
            cls: ClassVar[type[Any]]
            result: ClassVar[Any]

            @classmethod
            def from_crawler(cls, crawler):
                return cls(crawler=crawler)

            def __init__(self, crawler, **kwargs: Any):
                super().__init__(**kwargs)
                self.crawler = crawler

            async def start(self):
                MySpider.result = crawler.get_item_pipeline(MySpider.cls)
                return
                yield

        settings = {
            **BASE_SETTINGS,
            "ITEM_PIPELINES": {
                TrackingItemPipeline: 0,
            },
        }

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = TrackingItemPipeline
        await crawler.crawl_async()
        assert len(TrackingItemPipeline.instances) == 1
        assert MySpider.result == TrackingItemPipeline.instances[-1]

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = DefaultSpider
        await crawler.crawl_async()
        assert MySpider.result is None

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ParentItemPipeline
        await crawler.crawl_async()
        assert MySpider.result == TrackingItemPipeline.instances[-1]

        class ChildItemPipeline(TrackingItemPipeline):
            pass

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ChildItemPipeline
        await crawler.crawl_async()
        assert MySpider.result is None

    def test_get_item_pipeline_not_crawling(self) -> None:
        crawler = get_raw_crawler(settings_dict=BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            crawler.get_item_pipeline(DefaultSpider)

    @coroutine_test
    async def test_get_item_pipeline_no_engine(self) -> None:
        class MySpider(Spider):
            name = "myspider"

            @classmethod
            def from_crawler(cls, crawler):
                crawler.get_item_pipeline(DefaultSpider)

        crawler = get_raw_crawler(MySpider, BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            await crawler.crawl_async()

    @coroutine_test
    async def test_get_spider_middleware(self) -> None:
        class ParentSpiderMiddleware:
            pass

        class TrackingSpiderMiddleware(ParentSpiderMiddleware):
            instances: ClassVar[list[TrackingSpiderMiddleware]] = []

            def __init__(self):
                TrackingSpiderMiddleware.instances.append(self)

        class MySpider(Spider):
            name = "myspider"
            cls: ClassVar[type[Any]]
            result: ClassVar[Any]

            @classmethod
            def from_crawler(cls, crawler):
                return cls(crawler=crawler)

            def __init__(self, crawler, **kwargs: Any):
                super().__init__(**kwargs)
                self.crawler = crawler

            async def start(self):
                MySpider.result = crawler.get_spider_middleware(MySpider.cls)
                return
                yield

        settings = {
            **BASE_SETTINGS,
            "SPIDER_MIDDLEWARES": {
                TrackingSpiderMiddleware: 0,
            },
        }

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = TrackingSpiderMiddleware
        await crawler.crawl_async()
        assert len(TrackingSpiderMiddleware.instances) == 1
        assert MySpider.result == TrackingSpiderMiddleware.instances[-1]

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = DefaultSpider
        await crawler.crawl_async()
        assert MySpider.result is None

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ParentSpiderMiddleware
        await crawler.crawl_async()
        assert MySpider.result == TrackingSpiderMiddleware.instances[-1]

        class ChildSpiderMiddleware(TrackingSpiderMiddleware):
            pass

        crawler = get_raw_crawler(MySpider, settings)
        MySpider.cls = ChildSpiderMiddleware
        await crawler.crawl_async()
        assert MySpider.result is None

    def test_get_spider_middleware_not_crawling(self) -> None:
        crawler = get_raw_crawler(settings_dict=BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            crawler.get_spider_middleware(DefaultSpider)

    @coroutine_test
    async def test_get_spider_middleware_no_engine(self) -> None:
        class MySpider(Spider):
            name = "myspider"

            @classmethod
            def from_crawler(cls, crawler):
                crawler.get_spider_middleware(DefaultSpider)

        crawler = get_raw_crawler(MySpider, BASE_SETTINGS)
        with pytest.raises(RuntimeError):
            await crawler.crawl_async()


class TestSpiderSettings:
    def test_spider_custom_settings(self) -> None:
        class MySpider(scrapy.Spider):
            name = "spider"
            custom_settings = {"AUTOTHROTTLE_ENABLED": True}

        crawler = get_crawler(MySpider)
        enabled_exts = [e.__class__ for e in crawler.extensions.middlewares]
        assert AutoThrottle in enabled_exts


class TestCrawlerLogging:
    def test_no_root_handler_installed(self) -> None:
        handler = get_scrapy_root_handler()
        if handler is not None:
            logging.root.removeHandler(handler)

        class MySpider(scrapy.Spider):
            name = "spider"

        get_crawler(MySpider)
        assert get_scrapy_root_handler() is None

    @coroutine_test
    async def test_spider_custom_settings_log_level(self, tmp_path: Path) -> None:
        log_file = Path(tmp_path, "log.txt")
        log_file.write_text("previous message\n", encoding="utf-8")

        info_count = None

        class MySpider(scrapy.Spider):
            name = "spider"
            custom_settings = {
                "LOG_LEVEL": "INFO",
                "LOG_FILE": str(log_file),
            }

            async def start(self):
                info_count_start = crawler.stats.get_value("log_count/INFO")
                logging.debug("debug message")  # noqa: LOG015
                logging.info("info message")  # noqa: LOG015
                logging.warning("warning message")  # noqa: LOG015
                logging.error("error message")  # noqa: LOG015
                nonlocal info_count
                info_count = (
                    crawler.stats.get_value("log_count/INFO") - info_count_start
                )
                return
                yield

        try:
            configure_logging()
            handler = get_scrapy_root_handler()
            assert handler is not None
            assert handler.level == logging.DEBUG
            crawler = get_crawler(MySpider)
            handler = get_scrapy_root_handler()
            assert handler is not None
            assert handler.level == logging.INFO
            await crawler.crawl_async()
        finally:
            _uninstall_scrapy_root_handler()

        logged = log_file.read_text(encoding="utf-8")

        assert "previous message" in logged
        assert "debug message" not in logged
        assert "info message" in logged
        assert "warning message" in logged
        assert "error message" in logged
        assert crawler.stats.get_value("log_count/ERROR") == 1
        assert crawler.stats.get_value("log_count/WARNING") == 1
        assert info_count == 1
        assert crawler.stats.get_value("log_count/DEBUG", 0) == 0

    def test_spider_custom_settings_log_append(self, tmp_path: Path) -> None:
        log_file = Path(tmp_path, "log.txt")
        log_file.write_text("previous message\n", encoding="utf-8")

        class MySpider(scrapy.Spider):
            name = "spider"
            custom_settings = {
                "LOG_FILE": str(log_file),
                "LOG_FILE_APPEND": False,
            }

        try:
            configure_logging()
            get_crawler(MySpider)
            logging.debug("debug message")  # noqa: LOG015
        finally:
            _uninstall_scrapy_root_handler()

        logged = log_file.read_text(encoding="utf-8")

        assert "previous message" not in logged
        assert "debug message" in logged


class TestAsyncCrawlerProcessReactorlessHelpers:
    """Unit tests for the reactorless shutdown helpers of AsyncCrawlerProcess.

    These cover defensive branches that guard against shutdown races and that
    are not reachable through a full process run.
    """

    @staticmethod
    def _bare_process(
        monkeypatch: pytest.MonkeyPatch,
    ) -> tuple[AsyncCrawlerProcess, list[Any]]:
        # AsyncCrawlerProcess.__init__ has global side effects (it installs a
        # reactor import hook and an asyncio event loop), so build a bare
        # instance and set only the attributes these helpers read. The shutdown
        # handlers installed by these helpers are recorded for assertions
        # instead of touching the real process-wide signal handlers.
        installed_handlers: list[Any] = []
        monkeypatch.setattr(
            "scrapy.crawler.install_shutdown_handlers",
            lambda handler, *args, **kwargs: installed_handlers.append(handler),
        )
        return AsyncCrawlerProcess.__new__(AsyncCrawlerProcess), installed_handlers

    @staticmethod
    def _run_in_thread(target: Callable[[], None]) -> None:
        # Run target in a dedicated thread so its event loop is not nested
        # inside the event loop that may already be running the test session.
        thread = threading.Thread(target=target)
        thread.start()
        thread.join()

    def test_signal_shutdown_reactorless_without_loop(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, installed_handlers = self._bare_process(monkeypatch)
        process._reactorless_loop = None
        # No loop to schedule the shutdown task on, so it returns early, but it
        # must still escalate the handler so a second signal forces a kill.
        process._signal_shutdown_reactorless(signal.SIGINT, None)
        assert installed_handlers == [process._signal_kill_reactorless]

    def test_signal_kill_reactorless_without_loop(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, installed_handlers = self._bare_process(monkeypatch)
        process._reactorless_loop = None
        process._reactorless_main_task = None
        # No loop to cancel the main task on, so it returns early, but it must
        # still ignore any further signals.
        process._signal_kill_reactorless(signal.SIGINT, None)
        assert installed_handlers == [signal.SIG_IGN]

    def test_signal_kill_reactorless_without_main_task(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, installed_handlers = self._bare_process(monkeypatch)
        loop = MagicMock()
        process._reactorless_loop = loop
        process._reactorless_main_task = None
        # No main task to cancel, only the log message is scheduled.
        process._signal_kill_reactorless(signal.SIGINT, None)
        assert installed_handlers == [signal.SIG_IGN]
        loop.call_soon_threadsafe.assert_called_once_with(
            process._log_kill, signal.SIGINT
        )

    def test_signal_shutdown_reactorless_does_not_log_inline(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, _ = self._bare_process(monkeypatch)
        loop = MagicMock()
        process._reactorless_loop = loop
        log_calls: list[int] = []
        monkeypatch.setattr(process, "_log_shutdown", log_calls.append)
        # A signal can land while the interrupted code is itself writing to
        # the log stream, so logging from the handler would reenter that
        # write; it must be scheduled on the loop instead of called directly.
        process._signal_shutdown_reactorless(signal.SIGINT, None)
        assert log_calls == []
        loop.call_soon_threadsafe.assert_any_call(process._log_shutdown, signal.SIGINT)

    def test_signal_kill_reactorless_does_not_log_inline(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, _ = self._bare_process(monkeypatch)
        loop = MagicMock()
        process._reactorless_loop = loop
        process._reactorless_main_task = MagicMock()
        log_calls: list[int] = []
        monkeypatch.setattr(process, "_log_kill", log_calls.append)
        process._signal_kill_reactorless(signal.SIGINT, None)
        assert log_calls == []
        loop.call_soon_threadsafe.assert_any_call(process._log_kill, signal.SIGINT)

    def test_shutdown_graceful_reactorless_main_task_already_done(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, _ = self._bare_process(monkeypatch)
        process._stop_after_crawl = False

        async def noop() -> None:
            return None

        monkeypatch.setattr(process, "stop", noop)
        monkeypatch.setattr(process, "join", noop)

        def run() -> None:
            loop = asyncio.new_event_loop()
            try:
                main_task: asyncio.Future[None] = loop.create_future()
                main_task.set_result(None)
                process._reactorless_main_task = main_task
                # The main task is already done, so it is not cancelled.
                loop.run_until_complete(process._shutdown_graceful_reactorless())
                assert not main_task.cancelled()
            finally:
                loop.close()

        self._run_in_thread(run)

    def test_create_shutdown_task_closed_loop(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, _ = self._bare_process(monkeypatch)
        loop = asyncio.new_event_loop()
        loop.close()
        process._reactorless_loop = loop
        process._stop_after_crawl = True
        # create_task() raises RuntimeError on a closed loop; the coroutine
        # must be closed instead of leaking.
        process._create_shutdown_task()

    def test_cancel_all_tasks_logs_task_exception(self) -> None:
        contexts: list[dict[str, Any]] = []
        task_was_cancelled: list[bool] = []

        def run() -> None:
            loop = asyncio.new_event_loop()
            loop.set_exception_handler(lambda _loop, context: contexts.append(context))

            async def fail_on_cancel() -> None:
                try:
                    await asyncio.sleep(10)
                except asyncio.CancelledError:
                    raise RuntimeError("boom")

            try:
                task = loop.create_task(fail_on_cancel())
                # Let the task start and suspend on the sleep so the
                # cancellation is raised inside its body and turned into a
                # RuntimeError rather than cancelling the task cleanly.
                loop.run_until_complete(asyncio.sleep(0))
                AsyncCrawlerProcess._cancel_all_tasks(loop)
                task_was_cancelled.append(task.cancelled())
            finally:
                loop.close()

        self._run_in_thread(run)

        # The task raised instead of being cancelled, so its exception is
        # reported to the loop exception handler.
        assert task_was_cancelled == [False]
        assert any(
            context.get("message")
            == "unhandled exception during AsyncCrawlerProcess shutdown"
            for context in contexts
        )


class TestCrawlerProcessBaseSignalHandlers:
    """Unit tests for the Twisted-reactor shutdown handlers shared by
    CrawlerProcess and AsyncCrawlerProcess.
    """

    @staticmethod
    def _bare_process(
        monkeypatch: pytest.MonkeyPatch,
    ) -> tuple[CrawlerProcess, list[Any]]:
        installed_handlers: list[Any] = []
        monkeypatch.setattr(
            "scrapy.crawler.install_shutdown_handlers",
            lambda handler, *args, **kwargs: installed_handlers.append(handler),
        )
        return CrawlerProcess.__new__(CrawlerProcess), installed_handlers

    def test_signal_shutdown_does_not_log_inline(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, _ = self._bare_process(monkeypatch)
        reactor = MagicMock()
        # No reactor may be installed yet in the test process, so
        # twisted.internet may not have a "reactor" attribute to patch over.
        monkeypatch.setattr("twisted.internet.reactor", reactor, raising=False)
        log_calls: list[int] = []
        monkeypatch.setattr(process, "_log_shutdown", log_calls.append)
        # A signal can land while the interrupted code is itself writing to
        # the log stream, so logging from the handler would reenter that
        # write and raise, aborting the handler before it schedules the
        # graceful stop; it must be scheduled on the reactor instead.
        process._signal_shutdown(signal.SIGINT, None)
        assert log_calls == []
        reactor.callFromThread.assert_any_call(process._log_shutdown, signal.SIGINT)
        reactor.callFromThread.assert_any_call(process._graceful_stop_reactor)

    def test_signal_kill_does_not_log_inline(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        process, _ = self._bare_process(monkeypatch)
        reactor = MagicMock()
        monkeypatch.setattr("twisted.internet.reactor", reactor, raising=False)
        log_calls: list[int] = []
        monkeypatch.setattr(process, "_log_kill", log_calls.append)
        process._signal_kill(signal.SIGINT, None)
        assert log_calls == []
        reactor.callFromThread.assert_any_call(process._log_kill, signal.SIGINT)
        reactor.callFromThread.assert_any_call(process._stop_reactor)


@pytest.mark.parametrize(
    ("settings", "items"),
    [
        ({}, default_settings.LOG_VERSIONS),
        ({"LOG_VERSIONS": ["itemadapter"]}, ["itemadapter"]),
        ({"LOG_VERSIONS": []}, None),
    ],
)
def test_log_scrapy_info(
    settings: dict[str, Any], items: list[str] | None, caplog: pytest.LogCaptureFixture
) -> None:
    with caplog.at_level("INFO"):
        CrawlerProcess(settings, install_root_handler=False)
    assert (
        caplog.records[0].getMessage()
        == f"Scrapy {scrapy.__version__} started (bot: scrapybot)"
    ), repr(caplog.records[0].msg)
    if not items:
        assert len(caplog.records) == 1
        return
    version_string = caplog.records[1].getMessage()
    expected_items_pattern = "',\n '".join(
        f"{item}': '[^']+('\n +'[^']+)*" for item in items
    )
    assert re.search(r"^Versions:\n{'" + expected_items_pattern + "'}$", version_string)


@coroutine_test
async def test_deprecated_crawler_stop() -> None:
    crawler = get_crawler(DefaultSpider)
    d = crawler.crawl()
    await maybe_deferred_to_future(d)
    with pytest.warns(
        ScrapyDeprecationWarning, match=r"Crawler.stop\(\) is deprecated"
    ):
        await maybe_deferred_to_future(crawler.stop())



from __future__ import annotations

import re
from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

import attr
from itemadapter import ItemAdapter
from pydispatch import dispatcher
from twisted.internet import defer

from scrapy import signals
from scrapy.http import Headers, Request, Response
from scrapy.item import Field, Item
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Spider
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.signal import disconnect_all
from scrapy.utils.test import get_crawler

if TYPE_CHECKING:
    from twisted.python.failure import Failure

    from tests.mockserver.http import MockServer


class MyItem(Item):
    name = Field()
    url = Field()
    price = Field()


@attr.s
class AttrsItem:
    name = attr.ib(default="")
    url = attr.ib(default="")
    price = attr.ib(default=0)


@dataclass
class DataClassItem:
    name: str = ""
    url: str = ""
    price: int = 0


class MySpider(Spider):
    name = "scrapytest.org"

    itemurl_re = re.compile(r"item\d+.html")
    name_re = re.compile(r"<h1>(.*?)</h1>", re.MULTILINE)
    price_re = re.compile(r">Price: \$(.*?)<", re.MULTILINE)

    item_cls: type = MyItem

    def parse(self, response):
        xlink = LinkExtractor()
        itemre = re.compile(self.itemurl_re)
        for link in xlink.extract_links(response):
            if itemre.search(link.url):
                yield Request(url=link.url, callback=self.parse_item)

    def parse_item(self, response):
        adapter = ItemAdapter(self.item_cls())
        m = self.name_re.search(response.text)
        if m:
            adapter["name"] = m.group(1)
        adapter["url"] = response.url
        m = self.price_re.search(response.text)
        if m:
            adapter["price"] = m.group(1)
        return adapter.item


class DictItemsSpider(MySpider):
    item_cls = dict


class AttrsItemsSpider(MySpider):
    item_cls = AttrsItem


class DataClassItemsSpider(MySpider):
    item_cls = DataClassItem


class CrawlerRun:
    """A class to run the crawler and keep track of events occurred"""

    def __init__(self, spider_class: type[Spider]):
        self.respplug: list[tuple[Response, Spider]] = []
        self.reqplug: list[tuple[Request, Spider]] = []
        self.reqdropped: list[tuple[Request, Spider]] = []
        self.reqreached: list[tuple[Request, Spider]] = []
        self.itemerror: list[tuple[Any, Response, Spider, Failure]] = []
        self.itemresp: list[tuple[Any, Response]] = []
        self.headers: dict[Request, Headers] = {}
        self.bytes: defaultdict[Request, list[bytes]] = defaultdict(list)
        self.signals_caught: dict[Any, dict[str, Any]] = {}
        self.spider_class = spider_class

    async def run(self, mockserver: MockServer) -> None:
        self.mockserver = mockserver

        start_urls = [
            self.geturl("/static/"),
            self.geturl("/redirect"),
            self.geturl("/redirect"),  # duplicate
            self.geturl("/numbers"),
        ]

        for name, signal in vars(signals).items():
            if not name.startswith("_"):
                dispatcher.connect(self.record_signal, signal)

        self.crawler = get_crawler(self.spider_class)
        self.crawler.signals.connect(self.item_scraped, signals.item_scraped)
        self.crawler.signals.connect(self.item_error, signals.item_error)
        self.crawler.signals.connect(self.headers_received, signals.headers_received)
        self.crawler.signals.connect(self.bytes_received, signals.bytes_received)
        self.crawler.signals.connect(self.request_scheduled, signals.request_scheduled)
        self.crawler.signals.connect(self.request_dropped, signals.request_dropped)
        self.crawler.signals.connect(
            self.request_reached, signals.request_reached_downloader
        )
        self.crawler.signals.connect(
            self.response_downloaded, signals.response_downloaded
        )
        self.crawler.crawl(start_urls=start_urls)

        self.deferred: defer.Deferred[None] = defer.Deferred()
        dispatcher.connect(self.stop, signals.engine_stopped)
        await maybe_deferred_to_future(self.deferred)

    async def stop(self):
        for name, signal in vars(signals).items():
            if not name.startswith("_"):
                disconnect_all(signal)
        self.deferred.callback(None)
        await self.crawler.stop_async()

    def geturl(self, path: str) -> str:
        return self.mockserver.url(path)

    def getpath(self, url: str) -> str:
        u = urlparse(url)
        return u.path

    def item_error(
        self, item: Any, response: Response, spider: Spider, failure: Failure
    ) -> None:
        self.itemerror.append((item, response, spider, failure))

    def item_scraped(self, item: Any, spider: Spider, response: Response) -> None:
        self.itemresp.append((item, response))

    def headers_received(
        self, headers: Headers, body_length: int, request: Request, spider: Spider
    ) -> None:
        self.headers[request] = headers

    def bytes_received(self, data: bytes, request: Request, spider: Spider) -> None:
        self.bytes[request].append(data)

    def request_scheduled(self, request: Request, spider: Spider) -> None:
        self.reqplug.append((request, spider))

    def request_reached(self, request: Request, spider: Spider) -> None:
        self.reqreached.append((request, spider))

    def request_dropped(self, request: Request, spider: Spider) -> None:
        self.reqdropped.append((request, spider))

    def response_downloaded(self, response: Response, spider: Spider) -> None:
        self.respplug.append((response, spider))

    def record_signal(self, *args: Any, **kwargs: Any) -> None:
        """Record a signal and its parameters"""
        signalargs = kwargs.copy()
        sig = signalargs.pop("signal")
        signalargs.pop("sender", None)
        self.signals_caught[sig] = signalargs



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from scrapy.signals import request_left_downloader
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler
from tests.utils.decorators import inline_callbacks_test

if TYPE_CHECKING:
    from scrapy import Request
    from scrapy.crawler import Crawler
    from tests.mockserver.http import MockServer


class SignalCatcherSpider(Spider):
    name = "signal_catcher"

    def __init__(self, crawler: Crawler, url: str, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        crawler.signals.connect(self.on_request_left, signal=request_left_downloader)
        self.caught_times = 0
        self.start_urls = [url]

    @classmethod
    def from_crawler(
        cls, crawler: Crawler, *args: Any, **kwargs: Any
    ) -> SignalCatcherSpider:
        return cls(crawler, *args, **kwargs)

    def on_request_left(self, request: Request, spider: Spider) -> None:
        self.caught_times += 1


@inline_callbacks_test
def test_success(mockserver: MockServer):
    crawler = get_crawler(SignalCatcherSpider)
    yield crawler.crawl(mockserver.url("/status?n=200"))
    assert isinstance(crawler.spider, SignalCatcherSpider)
    assert crawler.spider.caught_times == 1


@inline_callbacks_test
def test_timeout(mockserver: MockServer):
    crawler = get_crawler(SignalCatcherSpider, {"DOWNLOAD_TIMEOUT": 0.1})
    yield crawler.crawl(mockserver.url("/delay?n=0.2"))
    assert isinstance(crawler.spider, SignalCatcherSpider)
    assert crawler.spider.caught_times == 1


@inline_callbacks_test
def test_disconnect(mockserver: MockServer):
    crawler = get_crawler(SignalCatcherSpider)
    yield crawler.crawl(mockserver.url("/drop"))
    assert isinstance(crawler.spider, SignalCatcherSpider)
    assert crawler.spider.caught_times == 1


@inline_callbacks_test
def test_noconnect():
    crawler = get_crawler(SignalCatcherSpider)
    yield crawler.crawl("http://thereisdefinetelynosuchdomain.com")
    assert isinstance(crawler.spider, SignalCatcherSpider)
    assert crawler.spider.caught_times == 1



"""
This module contains some assorted functions used in tests
"""

from __future__ import annotations

import asyncio
import os
import warnings
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, cast

from twisted.web.client import Agent

from scrapy.crawler import AsyncCrawlerRunner, CrawlerRunner, CrawlerRunnerBase
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.utils.reactor import is_asyncio_reactor_installed, is_reactor_installed
from scrapy.utils.spider import DefaultSpider

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from twisted.internet.defer import Deferred
    from twisted.web.client import Response as TxResponse

    from scrapy import Spider
    from scrapy.crawler import Crawler


_T = TypeVar("_T")


def get_reactor_settings() -> dict[str, Any]:
    """Return a settings dict that works with the installed reactor.

    ``Crawler._apply_settings()`` checks that the installed reactor matches the
    settings, so tests that run the crawler in the current process may need to
    pass a correct :setting:`TWISTED_REACTOR` setting value when creating it.
    """
    settings: dict[str, Any] = {}
    if is_reactor_installed():
        if not is_asyncio_reactor_installed():
            settings["TWISTED_REACTOR"] = None
    else:
        # We are either running Scrapy tests for the reactorless mode, or
        # running some 3rd-party library tests for the reactorless mode, or
        # running some 3rd-party library tests without initializing a reactor
        # properly. The first two cases are fine, but we cannot distinguish the
        # last one from them.
        settings["TWISTED_REACTOR_ENABLED"] = False
        settings["DOWNLOAD_HANDLERS"] = {
            "ftp": None,
            "http": "scrapy.core.downloader.handlers._httpx.HttpxDownloadHandler",
            "https": "scrapy.core.downloader.handlers._httpx.HttpxDownloadHandler",
        }
    return settings


def get_crawler(
    spidercls: type[Spider] | None = None,
    settings_dict: dict[str, Any] | None = None,
    prevent_warnings: bool = True,
) -> Crawler:
    """Return an unconfigured Crawler object. If settings_dict is given, it
    will be used to populate the crawler settings with a project level
    priority.
    """
    # When needed, useful settings can be added here, e.g. ones that prevent
    # deprecation warnings.
    settings: dict[str, Any] = {
        "TELNETCONSOLE_ENABLED": False,
        **get_reactor_settings(),
        **(settings_dict or {}),
    }
    runner: CrawlerRunnerBase
    if is_reactor_installed():
        runner = CrawlerRunner(settings)
    else:
        runner = AsyncCrawlerRunner(settings)
    crawler = runner.create_crawler(spidercls or DefaultSpider)
    crawler._apply_settings()
    return crawler


def get_pythonpath() -> str:
    """Return a PYTHONPATH suitable to use in processes so that they find this
    installation of Scrapy"""
    scrapy_path = import_module("scrapy").__path__[0]
    return str(Path(scrapy_path).parent) + os.pathsep + os.environ.get("PYTHONPATH", "")


def get_testenv() -> dict[str, str]:
    """Return a OS environment dict suitable to fork processes that need to import
    this installation of Scrapy, instead of a system installed one.
    """
    env = os.environ.copy()
    env["PYTHONPATH"] = get_pythonpath()
    return env


def get_from_asyncio_queue(value: _T) -> Awaitable[_T]:
    q: asyncio.Queue[_T] = asyncio.Queue()
    getter = q.get()
    q.put_nowait(value)
    return getter


def get_web_client_agent_req(url: str) -> Deferred[TxResponse]:  # pragma: no cover
    warnings.warn(
        "The get_web_client_agent_req() function is deprecated"
        " and will be removed in a future version of Scrapy.",
        category=ScrapyDeprecationWarning,
        stacklevel=2,
    )

    from twisted.internet import reactor

    agent = Agent(reactor)
    return cast("Deferred[TxResponse]", agent.request(b"GET", url.encode("utf-8")))



import logging
import sys

import scrapy
from scrapy.crawler import AsyncCrawlerProcess
from scrapy.utils.reactorless import ReactorImportHook

logger = logging.getLogger(__name__)


class NoRequestsSpider(scrapy.Spider):
    name = "no_request"

    async def start(self):
        return
        yield


process = AsyncCrawlerProcess(settings={"TWISTED_REACTOR_ENABLED": False})

process.crawl(NoRequestsSpider)
process.start()

hook_count = sum(1 for finder in sys.meta_path if isinstance(finder, ReactorImportHook))
logger.info(f"Hooks in sys.meta_path after start(): {hook_count}")

import twisted.internet.reactor  # noqa: E402,F401,TID253

logger.info("Reactor imported after start()")



from __future__ import annotations

import pytest

from scrapy.exceptions import ScrapyDeprecationWarning


def test_deprecated_concurrent_requests_per_ip_attribute() -> None:
    with pytest.warns(
        ScrapyDeprecationWarning,
        match=r"scrapy\.settings\.default_settings\.CONCURRENT_REQUESTS_PER_IP attribute is deprecated",
    ):
        from scrapy.settings.default_settings import (  # noqa: PLC0415
            CONCURRENT_REQUESTS_PER_IP,
        )

    assert CONCURRENT_REQUESTS_PER_IP is not None
    assert isinstance(CONCURRENT_REQUESTS_PER_IP, int)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest

from scrapy.robotstxt import (
    ProtegoRobotParser,
    PythonRobotParser,
    RerpRobotParser,
    RobotParser,
    decode_robotstxt,
)
from scrapy.utils._deps_compat import STDLIB_IMPROVED_ROBOTFILEPARSER
from scrapy.utils.misc import build_from_crawler
from tests.utils.robotstxt import rerp_available

if TYPE_CHECKING:
    # typing.Self requires Python 3.11
    from typing_extensions import Self

    from scrapy.crawler import Crawler

# The parser backends only use the crawler to get the spider to log with.
NO_CRAWLER = cast("Crawler", None)


class BaseRobotParserTest:
    parser_cls: type[RobotParser]

    def _setUp(self, parser_cls: type[RobotParser]) -> None:
        self.parser_cls = parser_cls

    def _parse(self, robotstxt_body: bytes) -> RobotParser:
        return build_from_crawler(self.parser_cls, NO_CRAWLER, robotstxt_body)

    def test_allowed(self):
        robotstxt_robotstxt_body = (
            b"User-agent: * \nDisallow: /disallowed \nAllow: /allowed \nCrawl-delay: 10"
        )
        rp = self._parse(robotstxt_robotstxt_body)
        assert rp.allowed("https://www.site.local/allowed", "*")
        assert not rp.allowed("https://www.site.local/disallowed", "*")

    def test_allowed_wildcards(self) -> None:
        robotstxt_robotstxt_body = b"""User-agent: first
                                Disallow: /disallowed/*/end$

                                User-agent: second
                                Allow: /*allowed
                                Disallow: /
                                """
        rp = self._parse(robotstxt_robotstxt_body)

        assert rp.allowed("https://www.site.local/disallowed", "first")
        assert not rp.allowed("https://www.site.local/disallowed/xyz/end", "first")
        assert not rp.allowed("https://www.site.local/disallowed/abc/end", "first")
        assert rp.allowed("https://www.site.local/disallowed/xyz/endinglater", "first")

        assert rp.allowed("https://www.site.local/allowed", "second")
        assert rp.allowed("https://www.site.local/is_still_allowed", "second")
        assert rp.allowed("https://www.site.local/is_allowed_too", "second")

    def test_length_based_precedence(self) -> None:
        robotstxt_robotstxt_body = b"User-agent: * \nDisallow: / \nAllow: /page"
        rp = self._parse(robotstxt_robotstxt_body)
        assert rp.allowed("https://www.site.local/page", "*")

    def test_order_based_precedence(self) -> None:
        robotstxt_robotstxt_body = b"User-agent: * \nDisallow: / \nAllow: /page"
        rp = self._parse(robotstxt_robotstxt_body)
        assert not rp.allowed("https://www.site.local/page", "*")

    def test_empty_response(self):
        """empty response should equal 'allow all'"""
        rp = self._parse(b"")
        assert rp.allowed("https://site.local/", "*")
        assert rp.allowed("https://site.local/", "chrome")
        assert rp.allowed("https://site.local/index.html", "*")
        assert rp.allowed("https://site.local/disallowed", "*")

    def test_garbage_response(self):
        """garbage response should be discarded, equal 'allow all'"""
        robotstxt_robotstxt_body = b"GIF89a\xd3\x00\xfe\x00\xa2"
        rp = self._parse(robotstxt_robotstxt_body)
        assert rp.allowed("https://site.local/", "*")
        assert rp.allowed("https://site.local/", "chrome")
        assert rp.allowed("https://site.local/index.html", "*")
        assert rp.allowed("https://site.local/disallowed", "*")

    def test_crawl_delay(self):
        robotstxt_body = b"User-agent: *\nDisallow: /private\nCrawl-delay: 10\n"
        rp = self._parse(robotstxt_body)
        assert rp.crawl_delay("*") == 10.0

    def test_crawl_delay_unset(self):
        robotstxt_body = b"User-agent: *\nDisallow: /private\n"
        rp = self._parse(robotstxt_body)
        assert rp.crawl_delay("*") is None

    def test_unicode_url_and_useragent(self):
        robotstxt_robotstxt_body = """
        User-Agent: *
        Disallow: /admin/
        Disallow: /static/
        # taken from https://en.wikipedia.org/robots.txt
        Disallow: /wiki/K%C3%A4ytt%C3%A4j%C3%A4:
        Disallow: /wiki/Käyttäjä:

        User-Agent: UnicödeBöt
        Disallow: /some/randome/page.html""".encode()
        rp = self._parse(robotstxt_robotstxt_body)
        assert rp.allowed("https://site.local/", "*")
        assert not rp.allowed("https://site.local/admin/", "*")
        assert not rp.allowed("https://site.local/static/", "*")
        assert rp.allowed("https://site.local/admin/", "UnicödeBöt")
        assert not rp.allowed("https://site.local/wiki/K%C3%A4ytt%C3%A4j%C3%A4:", "*")
        assert not rp.allowed("https://site.local/wiki/Käyttäjä:", "*")
        assert rp.allowed("https://site.local/some/randome/page.html", "*")
        assert not rp.allowed("https://site.local/some/randome/page.html", "UnicödeBöt")


class TestRobotParser:
    def test_crawl_delay_unsupported(self):
        class AllowAllRobotParser(RobotParser):
            @classmethod
            def from_crawler(cls, crawler: Crawler, robotstxt_body: bytes) -> Self:
                return cls()

            def allowed(self, url: str | bytes, user_agent: str | bytes) -> bool:
                return True

        rp = build_from_crawler(
            AllowAllRobotParser, NO_CRAWLER, b"User-agent: *\nCrawl-delay: 10\n"
        )
        assert rp.crawl_delay("*") is None


class TestDecodeRobotsTxt:
    def test_native_string_conversion(self):
        robotstxt_body = b"User-agent: *\nDisallow: /\n"
        decoded_content = decode_robotstxt(
            robotstxt_body, spider=None, to_native_str_type=True
        )
        assert decoded_content == "User-agent: *\nDisallow: /\n"

    def test_decode_utf8(self):
        robotstxt_body = b"User-agent: *\nDisallow: /\n"
        decoded_content = decode_robotstxt(robotstxt_body, spider=None)
        assert decoded_content == "User-agent: *\nDisallow: /\n"

    def test_decode_non_utf8(self):
        robotstxt_body = b"User-agent: *\n\xffDisallow: /\n"
        decoded_content = decode_robotstxt(robotstxt_body, spider=None)
        assert decoded_content == "User-agent: *\nDisallow: /\n"

    # UTF-8 BOM at the beginning of the file ignored
    def test_decode_utf8_bom(self):
        robotstxt_body = b"\xef\xbb\xbfUser-agent: *\nDisallow: /\n"
        decoded_content = decode_robotstxt(robotstxt_body, spider=None)
        assert decoded_content == "User-agent: *\nDisallow: /\n"


class TestPythonRobotParser(BaseRobotParserTest):
    def setup_method(self):
        super()._setUp(PythonRobotParser)

    @pytest.mark.skipif(
        not STDLIB_IMPROVED_ROBOTFILEPARSER,
        reason="RobotFileParser from this Python version does not support length based directives precedence.",
    )
    def test_length_based_precedence(self) -> None:
        super().test_length_based_precedence()

    @pytest.mark.skipif(
        STDLIB_IMPROVED_ROBOTFILEPARSER,
        reason="RobotFileParser from this Python version does not support order based directives precedence.",
    )
    def test_order_based_precedence(self) -> None:
        super().test_order_based_precedence()

    @pytest.mark.skipif(
        not STDLIB_IMPROVED_ROBOTFILEPARSER,
        reason="RobotFileParser from this Python version does not support wildcards.",
    )
    def test_allowed_wildcards(self) -> None:
        super().test_allowed_wildcards()


@pytest.mark.skipif(not rerp_available(), reason="Rerp parser is not installed")
class TestRerpRobotParser(BaseRobotParserTest):
    def setup_method(self):
        super()._setUp(RerpRobotParser)

    def test_length_based_precedence(self) -> None:
        pytest.skip("Rerp does not support length based directives precedence.")


class TestProtegoRobotParser(BaseRobotParserTest):
    def setup_method(self):
        super()._setUp(ProtegoRobotParser)

    def test_order_based_precedence(self) -> None:
        pytest.skip("Protego does not support order based directives precedence.")



"""
MemoryUsage extension

See documentation in docs/topics/extensions.rst
"""

from __future__ import annotations

import logging
import socket
import sys
import warnings
from importlib import import_module
from pprint import pformat
from typing import TYPE_CHECKING

from scrapy import signals
from scrapy.exceptions import NotConfigured, ScrapyDeprecationWarning
from scrapy.utils.asyncio import AsyncioLoopingCall, create_looping_call
from scrapy.utils.defer import _schedule_coro
from scrapy.utils.engine import get_engine_status
from scrapy.utils.misc import build_from_crawler

if TYPE_CHECKING:
    from twisted.internet.task import LoopingCall

    # typing.Self requires Python 3.11
    from typing_extensions import Self

    from scrapy.crawler import Crawler
    from scrapy.statscollectors import StatsCollector


logger = logging.getLogger(__name__)


class MemoryUsage:
    def __init__(self, crawler: Crawler):
        if not crawler.settings.getbool("MEMUSAGE_ENABLED"):
            raise NotConfigured
        try:
            # stdlib's resource module is only available on unix platforms.
            self.resource = import_module("resource")
        except ImportError as exc:
            raise NotConfigured from exc

        self.crawler: Crawler = crawler
        self._stats: StatsCollector = crawler.stats
        self.warned: bool = False
        self.notify_mails: list[str] = crawler.settings.getlist("MEMUSAGE_NOTIFY_MAIL")
        if self.notify_mails:  # pragma: no cover
            from scrapy.mail import MailSender  # noqa: PLC0415

            warnings.warn(
                "The 'MEMUSAGE_NOTIFY_MAIL' setting is deprecated and will be removed "
                "in a future release. Please use the 'memusage_warning_reached' and 'spider_closed' "
                "signals to implement custom notifications.",
                category=ScrapyDeprecationWarning,
                stacklevel=2,
            )
            self.mail = build_from_crawler(MailSender, crawler)

        self.limit: int = crawler.settings.getint("MEMUSAGE_LIMIT_MB") * 1024 * 1024
        self.warning: int = crawler.settings.getint("MEMUSAGE_WARNING_MB") * 1024 * 1024
        self.check_interval: float = crawler.settings.getfloat(
            "MEMUSAGE_CHECK_INTERVAL_SECONDS"
        )
        crawler.signals.connect(self.engine_started, signal=signals.engine_started)
        crawler.signals.connect(self.engine_stopped, signal=signals.engine_stopped)

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        return cls(crawler)

    def get_virtual_size(self) -> int:
        size: int = self.resource.getrusage(self.resource.RUSAGE_SELF).ru_maxrss
        if sys.platform != "darwin":
            # on macOS ru_maxrss is in bytes, on Linux it is in KB
            size *= 1024
        return size

    def engine_started(self) -> None:
        self._stats.set_value("memusage/startup", self.get_virtual_size())
        self.tasks: list[AsyncioLoopingCall | LoopingCall] = []
        tsk = create_looping_call(self.update)
        self.tasks.append(tsk)
        tsk.start(self.check_interval, now=True)
        if self.limit:
            tsk = create_looping_call(self._check_limit)
            self.tasks.append(tsk)
            tsk.start(self.check_interval, now=True)
        if self.warning:
            tsk = create_looping_call(self._check_warning)
            self.tasks.append(tsk)
            tsk.start(self.check_interval, now=True)

    def engine_stopped(self) -> None:
        for tsk in self.tasks:
            if tsk.running:  # pragma: no branch
                tsk.stop()

    def update(self) -> None:
        self._stats.max_value("memusage/max", self.get_virtual_size())

    def _check_limit(self) -> None:
        peak_mem_usage = self.get_virtual_size()
        if peak_mem_usage > self.limit:
            self._stats.set_value("memusage/limit_reached", 1)
            mem = self.limit / 1024 / 1024
            logger.error(
                "Memory usage exceeded %(memusage)dMiB. Shutting down Scrapy...",
                {"memusage": mem},
                extra={"crawler": self.crawler},
            )
            if self.notify_mails:  # pragma: no cover
                subj = (
                    f"{self.crawler.settings['BOT_NAME']} terminated: "
                    f"memory usage exceeded {mem}MiB at {socket.gethostname()}"
                )
                self._send_report(self.notify_mails, subj)
                self._stats.set_value("memusage/limit_notified", 1)

            if self.crawler.engine.spider is not None:
                _schedule_coro(
                    self.crawler.engine.close_spider_async(reason="memusage_exceeded")
                )
            else:
                _schedule_coro(self.crawler.stop_async())
        else:
            logger.info(
                "Peak memory usage is %(virtualsize)dMiB",
                {"virtualsize": peak_mem_usage / 1024 / 1024},
            )

    def _check_warning(self) -> None:
        if self.warned:  # warn only once
            return
        if self.get_virtual_size() > self.warning:
            self._stats.set_value("memusage/warning_reached", 1)
            self.crawler.signals.send_catch_log(signal=signals.memusage_warning_reached)
            mem = self.warning / 1024 / 1024
            logger.warning(
                "Memory usage reached %(memusage)dMiB",
                {"memusage": mem},
                extra={"crawler": self.crawler},
            )
            if self.notify_mails:  # pragma: no cover
                subj = (
                    f"{self.crawler.settings['BOT_NAME']} warning: "
                    f"memory usage reached {mem}MiB at {socket.gethostname()}"
                )
                self._send_report(self.notify_mails, subj)
                self._stats.set_value("memusage/warning_notified", 1)
            self.warned = True

    def _send_report(self, rcpts: list[str], subject: str) -> None:  # pragma: no cover
        """send notification mail with some additional useful info"""
        s = f"Memory usage at engine startup : {self._stats.get_value('memusage/startup') / 1024 / 1024}M\r\n"
        s += f"Maximum memory usage          : {self._stats.get_value('memusage/max') / 1024 / 1024}M\r\n"
        s += f"Current memory usage          : {self.get_virtual_size() / 1024 / 1024}M\r\n"

        s += (
            "ENGINE STATUS ------------------------------------------------------- \r\n"
        )
        s += "\r\n"
        s += pformat(get_engine_status(self.crawler.engine))
        s += "\r\n"
        self.mail.send(rcpts, subject, s)



"""Fallbacks for hostnames that the ``idna`` package rejects but that resolve fine.

Twisted encodes hostnames with ``idna``, which enforces IDNA 2008 and rejects
emoji domains even in punycode (``xn--``) form
(https://github.com/scrapy/scrapy/issues/4330) and domains with underscores
(https://github.com/scrapy/scrapy/issues/3321). The standard library codec
implements IDNA 2003 and handles both, so :func:`_install_twisted_idna_fallbacks`
patches Twisted to fall back to it. Ideally Twisted would do this itself.
"""

from __future__ import annotations

import re

from twisted.internet import _idna

_original_idna_bytes = _idna._idnaBytes
_original_idna_text = _idna._idnaText

# Letters, digits and hyphens, plus the underscore that the idna package
# rejects. The stdlib codec passes any other ASCII through as well, e.g. the
# ":" that Twisted parses out of a bracketless IPv6 netloc, and Twisted must
# keep treating those as invalid hostnames instead of resolving them.
_HOSTNAME = re.compile(rb"[a-zA-Z0-9._\-]+\Z")


def _safe_hostname_bytes(hostname: str) -> bytes:
    """Punycode *hostname* with IDNA 2003, which accepts emoji labels and passes
    already-encoded (``xn--``) ones through.

    Raises UnicodeError if it cannot be encoded into a hostname at all.
    """
    try:
        encoded = hostname.encode("idna")
    except UnicodeError:
        # The stdlib codec also rejects some ASCII hostnames, e.g. overlong
        # labels; non-ASCII ones raise UnicodeError again here.
        encoded = hostname.encode("ascii")
    if not _HOSTNAME.match(encoded):
        raise UnicodeError(f"invalid hostname: {hostname}")
    return encoded


def _patched_idna_bytes(text: str) -> bytes:
    try:
        return _original_idna_bytes(text)
    except UnicodeError:  # parent class of idna.IDNAError
        return _safe_hostname_bytes(text)


def _patched_idna_text(octets: bytes) -> str:
    try:
        return _original_idna_text(octets)
    except UnicodeError:
        if not _HOSTNAME.match(octets):
            raise
        return octets.decode("idna")


def _install_twisted_idna_fallbacks() -> None:
    """Patch ``_idnaBytes()``/``_idnaText()`` wherever Twisted uses them for
    client connections.

    Each module binds them at import time, so every reference needs patching,
    not just ``twisted.internet._idna``. Names that are missing or already
    patched are left alone.
    """
    from twisted.internet import _resolver, _sslverify, endpoints  # noqa: PLC0415

    patches = (
        ("_idnaBytes", _original_idna_bytes, _patched_idna_bytes),
        ("_idnaText", _original_idna_text, _patched_idna_text),
    )
    for module in (_idna, _resolver, _sslverify, endpoints):
        for name, original, patched in patches:
            if getattr(module, name, None) is original:
                setattr(module, name, patched)



from __future__ import annotations

import socket
from pathlib import Path
from typing import TYPE_CHECKING, cast

from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.x509 import load_pem_x509_certificate
from OpenSSL import SSL
from twisted.internet.ssl import CertificateOptions, ContextFactory

from scrapy.core.downloader.tls import _TWISTED_VERSION_MAP
from scrapy.utils.python import to_bytes
from scrapy.utils.ssl import _get_cert_options_version_kwargs

if TYPE_CHECKING:
    from twisted.internet.interfaces import IOpenSSLContextFactory


def _free_port() -> int:
    # racy but should be fine for tests
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return cast("int", s.getsockname()[1])


def ssl_context_factory(
    keyfile: str = "keys/localhost.key",
    certfile: str = "keys/localhost.crt",
    *,
    cipher_string: str | None = None,
    tls_min_version: str | None = None,
    tls_max_version: str | None = None,
) -> IOpenSSLContextFactory:
    keyfile_path = Path(__file__).parent.parent / keyfile
    certfile_path = Path(__file__).parent.parent / certfile

    cert = load_pem_x509_certificate(certfile_path.read_bytes())
    key = load_pem_private_key(keyfile_path.read_bytes(), password=None)

    tls_min = _TWISTED_VERSION_MAP.get(tls_min_version) if tls_min_version else None
    tls_max = _TWISTED_VERSION_MAP.get(tls_max_version) if tls_max_version else None
    tls_version_kwargs = _get_cert_options_version_kwargs(tls_min, tls_max)
    # https://github.com/twisted/twisted/issues/12638
    factory: CertificateOptions = CertificateOptions(
        privateKey=key,  # type: ignore[arg-type]
        certificate=cert,  # type: ignore[arg-type]
        **tls_version_kwargs,
    )
    if cipher_string:
        ctx = factory.getContext()
        # disabling TLS1.3 because it unconditionally enables some strong ciphers
        ctx.set_options(SSL.OP_CIPHER_SERVER_PREFERENCE | SSL.OP_NO_TLSv1_3)
        ctx.set_cipher_list(to_bytes(cipher_string))
    return cast("ContextFactory", factory)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple, cast

from scrapy.core.downloader import Downloader
from scrapy.utils.httpobj import urlparse_cached

if TYPE_CHECKING:
    from scrapy.http import Request


class MockSlot(NamedTuple):
    active: list[Any]


class MockDownloader:
    def __init__(self) -> None:
        self.slots: dict[str, MockSlot] = {}

    def get_slot_key(self, request: Request) -> str:
        if Downloader.DOWNLOAD_SLOT in request.meta:
            return cast("str", request.meta[Downloader.DOWNLOAD_SLOT])

        return urlparse_cached(request).hostname or ""

    def increment(self, slot_key: str) -> None:
        slot = self.slots.setdefault(slot_key, MockSlot(active=[]))
        slot.active.append(1)

    def decrement(self, slot_key: str) -> None:
        slot = self.slots[slot_key]
        slot.active.pop()

    def close(self) -> None:
        pass



import scrapy
from tests.utils.cmdline import proc


def test_output() -> None:
    _, out, _ = proc("version")
    assert out.strip() == f"Scrapy {scrapy.__version__}"


def test_verbose_output() -> None:
    _, out, _ = proc("version", "-v")
    headers = [line.partition(":")[0].strip() for line in out.strip().splitlines()]
    assert headers == [
        "Scrapy",
        "lxml",
        "libxml2",
        "cssselect",
        "parsel",
        "w3lib",
        "Twisted",
        "Python",
        "pyOpenSSL",
        "cryptography",
        "Platform",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from scrapy.http import Request, Response
from scrapy.spidermiddlewares.start import StartSpiderMiddleware
from scrapy.spiders import Spider
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


@coroutine_test
async def test_async() -> None:
    crawler = get_crawler(Spider)
    mw = build_from_crawler(StartSpiderMiddleware, crawler)

    async def start() -> AsyncIterator[Request]:
        yield Request("data:,1")
        yield Request("data:,2", meta={"is_start_request": True})
        yield Request("data:,2", meta={"is_start_request": False})
        yield Request("data:,2", meta={"is_start_request": "foo"})

    result = [
        request.meta["is_start_request"] async for request in mw.process_start(start())
    ]
    assert result == [True, True, False, "foo"]


def test_spider_output_not_marked() -> None:
    # Requests from a non-None response (spider output) are not flagged.
    crawler = get_crawler(Spider)
    mw = build_from_crawler(StartSpiderMiddleware, crawler)
    response = Response("data:,")
    request = Request("data:,1")
    out = list(mw.process_spider_output(response, [request]))
    assert out == [request]
    assert "is_start_request" not in request.meta



from __future__ import annotations

import subprocess
import sys
from importlib.util import find_spec
from io import BytesIO
from types import ModuleType
from typing import TYPE_CHECKING

import pytest
from pexpect import EOF

from scrapy.utils.console import get_shell_embed_func, start_python_console
from scrapy.utils.test import get_testenv

if TYPE_CHECKING:
    from pathlib import Path

CONSOLE = """
from scrapy.utils.console import start_python_console

start_python_console(banner="SHELL-READY", shells=["ipython"])
"""

CONSOLE_IN_RUNNING_LOOP = """
import asyncio

from scrapy.utils.console import start_python_console


async def main():
    start_python_console(banner="SHELL-READY", shells=["ipython"])


asyncio.run(main())
"""


def test_get_shell_embed_func():
    shell = get_shell_embed_func(["invalid"])
    assert shell is None

    shell = get_shell_embed_func(["invalid", "python"])
    assert callable(shell)
    assert shell.__name__ == "_embed_standard_shell"


def test_get_shell_embed_func_known_shells():
    def embed(namespace: dict[str, object] | None = None, banner: str = "") -> None:
        pass

    known_shells = {"custom": lambda: embed}
    assert get_shell_embed_func(["custom"], known_shells) is embed
    assert get_shell_embed_func(["python"], known_shells) is None


def test_get_shell_embed_func_python():
    # the standard shell is always available
    shell = get_shell_embed_func(["python"])
    assert callable(shell)
    assert shell.__name__ == "_embed_standard_shell"


def test_get_shell_embed_func_bpython():
    pytest.importorskip("bpython")
    shell = get_shell_embed_func(["bpython"])
    assert callable(shell)
    assert shell.__name__ == "_embed_bpython_shell"


def test_embed_bpython_shell(monkeypatch: pytest.MonkeyPatch) -> None:
    # bpython is never the default shell, so a stand-in module takes the place
    # of the interactive session the IPython tests below use.
    calls: list[tuple[dict[str, object], str]] = []
    bpython = ModuleType("bpython")
    monkeypatch.setattr(
        bpython,
        "embed",
        lambda locals_, banner: calls.append((locals_, banner)),
        raising=False,
    )
    monkeypatch.setitem(sys.modules, "bpython", bpython)

    shell = get_shell_embed_func(["bpython"])
    assert shell is not None
    shell({"a": 1}, "SHELL-READY")
    assert calls == [({"a": 1}, "SHELL-READY")]


def test_get_shell_embed_func_ipython():
    pytest.importorskip("IPython")
    shell = get_shell_embed_func(["ipython"])
    assert shell is not None
    assert shell.__name__ == "_embed_ipython_shell"


def test_get_shell_embed_func_ptpython():
    pytest.importorskip("ptpython")
    shell = get_shell_embed_func(["ptpython"])
    assert shell is not None
    assert shell.__name__ == "_embed_ptpython_shell"


def test_get_shell_embed_func_default():
    # with no shells given, the first available shell in preference order
    # (ptpython, ipython, bpython, python) is returned; the standard shell
    # is always available, so the default is never None
    shell = get_shell_embed_func()
    assert shell is not None
    if find_spec("ptpython") is not None:
        expected = "_embed_ptpython_shell"
    elif find_spec("IPython") is not None:
        expected = "_embed_ipython_shell"
    elif find_spec("bpython") is not None:
        expected = "_embed_bpython_shell"
    else:
        expected = "_embed_standard_shell"
    assert shell.__name__ == expected


@pytest.mark.skipif(find_spec("IPython") is None, reason="IPython is not installed")
class TestIPythonShell:
    """Starting an IPython shell, with and without an asyncio event loop already
    running in the calling thread. The latter happens when inspect_response() is
    called from a spider callback while using the asyncio reactor."""

    @staticmethod
    def _env(tmp_path: Path) -> dict[str, str]:
        env = get_testenv()
        # Keep IPython away from the profile and history of the user running the tests.
        env["IPYTHONDIR"] = str(tmp_path)
        return env

    def test_simple_prompt(self, tmp_path: Path) -> None:
        """IPython falls back to its simple prompt, which needs no event loop,
        when stdin is not a TTY."""
        env = self._env(tmp_path)
        p = subprocess.run(
            [sys.executable, "-c", CONSOLE_IN_RUNNING_LOOP],
            check=False,
            capture_output=True,
            encoding="utf-8",
            timeout=60,
            env=env,
            stdin=subprocess.DEVNULL,
        )
        output = p.stdout + p.stderr
        assert "SHELL-READY" in output
        assert p.returncode == 0, output

    @pytest.mark.skipif(
        sys.platform == "win32", reason="requires a POSIX pseudo-terminal"
    )
    # The child of the pseudo-terminal fork execs right away, so the deadlocks
    # that Python warns about cannot happen.
    @pytest.mark.filterwarnings(
        "ignore:.*is multi-threaded, use of forkpty:DeprecationWarning"
    )
    @pytest.mark.parametrize(
        "script",
        [CONSOLE, CONSOLE_IN_RUNNING_LOOP],
        ids=["no_running_loop", "running_loop"],
    )
    def test_tty(self, tmp_path: Path, script: str) -> None:
        """IPython uses prompt_toolkit, which needs an event loop of its own,
        when stdin is a TTY."""
        # pexpect only defines spawn, which needs a pseudo-terminal, on POSIX.
        from pexpect import spawn  # noqa: PLC0415

        env = self._env(tmp_path)
        env.pop("IPY_TEST_SIMPLE_PROMPT", None)
        env["TERM"] = "xterm"
        logfile = BytesIO()
        p = spawn(
            sys.executable,
            ["-c", script],
            env=env,
            timeout=60,
        )
        p.logfile_read = logfile
        try:
            # Wait for the prompt, which prompt_toolkit draws once it is done
            # querying the terminal, before typing into it.
            p.expect(r"In \[")
            p.sendline("21*2")
            p.expect_exact("42")
            p.sendline("exit()")
            p.expect(EOF)
        finally:
            p.close()
        output = logfile.getvalue().decode()
        assert "Traceback" not in output
        assert p.exitstatus == 0, output


def test_start_python_console_exit(monkeypatch: pytest.MonkeyPatch) -> None:
    def embed(namespace: dict[str, object], banner: str) -> None:
        raise SystemExit

    monkeypatch.setattr(
        "scrapy.utils.console.get_shell_embed_func", lambda shells: embed
    )
    start_python_console()


def test_start_python_console_no_shell(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "scrapy.utils.console.get_shell_embed_func", lambda shells: None
    )
    start_python_console()



from __future__ import annotations

import sys
from unittest.mock import Mock, call

import pytest
from twisted.internet import defer

from scrapy.core.engine import ExecutionEngine
from scrapy.http import Request, Response
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.spider import DefaultSpider
from scrapy.utils.test import get_crawler
from tests.utils.decorators import coroutine_test


@pytest.fixture
def engine() -> ExecutionEngine:
    crawler = get_crawler(DefaultSpider)
    engine = ExecutionEngine(crawler, lambda _: None)
    engine.downloader.close()
    engine.downloader = Mock()
    engine._slot = Mock()
    engine._slot.inprogress = set()
    return engine


class TestEngineDownloadAsync:
    """Test cases for ExecutionEngine.download_async()."""

    @staticmethod
    async def _download(engine: ExecutionEngine, request: Request) -> Response:
        return await engine.download_async(request)

    @coroutine_test
    async def test_download_async_success(self, engine):
        """Test basic successful async download of a request."""
        request = Request("http://example.com")
        response = Response("http://example.com", body=b"test body")
        engine.spider = Mock()
        engine.downloader.fetch.return_value = defer.succeed(response)
        engine._slot.add_request = Mock()
        engine._slot.remove_request = Mock()

        result = await self._download(engine, request)
        assert result == response
        engine._slot.add_request.assert_called_once_with(request)
        engine._slot.remove_request.assert_called_once_with(request)
        engine.downloader.fetch.assert_called_once_with(request)

    @coroutine_test
    async def test_download_async_redirect(self, engine):
        """Test async download with a redirect request."""
        original_request = Request("http://example.com")
        redirect_request = Request("http://example.com/redirect")
        final_response = Response("http://example.com/redirect", body=b"redirected")

        # First call returns redirect request, second call returns final response
        engine.downloader.fetch.side_effect = [
            defer.succeed(redirect_request),
            defer.succeed(final_response),
        ]
        engine.spider = Mock()
        engine._slot.add_request = Mock()
        engine._slot.remove_request = Mock()

        result = await self._download(engine, original_request)
        assert result == final_response
        assert engine.downloader.fetch.call_count == 2
        engine._slot.add_request.assert_has_calls(
            [call(original_request), call(redirect_request)]
        )
        engine._slot.remove_request.assert_has_calls(
            [call(original_request), call(redirect_request)]
        )

    @coroutine_test
    async def test_download_async_many_redirects(self, engine):
        """A long chain of requests being replaced by new ones is handled
        iteratively, without hitting the recursion limit."""
        count = sys.getrecursionlimit() * 2
        requests = [Request(f"http://example.com/{i}") for i in range(count)]
        final_response = Response("http://example.com/final", body=b"done")
        engine.downloader.fetch.side_effect = [
            *(defer.succeed(request) for request in requests[1:]),
            defer.succeed(final_response),
        ]
        engine.spider = Mock()
        engine._slot.add_request = Mock()
        engine._slot.remove_request = Mock()

        result = await self._download(engine, requests[0])
        assert result == final_response
        assert engine.downloader.fetch.call_count == count

    @coroutine_test
    async def test_download_async_no_spider(self, engine):
        """Test async download attempt when no spider is available."""
        request = Request("http://example.com")
        engine.spider = None
        with pytest.raises(RuntimeError, match="No open spider to crawl:"):
            await self._download(engine, request)

    @coroutine_test
    async def test_download_async_failure(self, engine):
        """Test async download when the downloader raises an exception."""
        request = Request("http://example.com")
        error = RuntimeError("Download failed")
        engine.spider = Mock()
        engine.downloader.fetch.return_value = defer.fail(error)
        engine._slot.add_request = Mock()
        engine._slot.remove_request = Mock()

        with pytest.raises(RuntimeError, match="Download failed"):
            await self._download(engine, request)
        engine._slot.add_request.assert_called_once_with(request)
        engine._slot.remove_request.assert_called_once_with(request)

    @coroutine_test
    async def test_download_async_fetch_needs_spider(self, engine):
        engine._downloader_fetch_needs_spider = True
        request = Request("http://example.com")
        response = Response("http://example.com", body=b"test body")
        engine.spider = Mock()
        engine.downloader.fetch.return_value = defer.succeed(response)

        result = await self._download(engine, request)
        assert result == response
        engine.downloader.fetch.assert_called_once_with(request, engine.spider)


@pytest.mark.filterwarnings("ignore::scrapy.exceptions.ScrapyDeprecationWarning")
class TestEngineDownload(TestEngineDownloadAsync):
    """Test cases for ExecutionEngine.download()."""

    @staticmethod
    async def _download(engine: ExecutionEngine, request: Request) -> Response:
        return await maybe_deferred_to_future(engine.download(request))



"""
Depth Spider Middleware

See documentation in docs/topics/spider-middleware.rst
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from scrapy.spidermiddlewares.base import BaseSpiderMiddleware
from scrapy.utils.decorators import _warn_spider_arg

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterable

    # typing.Self requires Python 3.11
    from typing_extensions import Self

    from scrapy import Spider
    from scrapy.crawler import Crawler
    from scrapy.http import Request, Response
    from scrapy.statscollectors import StatsCollector


logger = logging.getLogger(__name__)


class DepthMiddleware(BaseSpiderMiddleware):
    """Track the depth of each request within the site being scraped, setting
    ``request.meta["depth"]`` to 0 when there is no value previously set
    (usually just the first request) and incrementing it by 1 otherwise.

    It can be used to limit the maximum depth to scrape, control request
    priority based on their depth, and things like that, through the
    :setting:`DEPTH_LIMIT`, :setting:`DEPTH_STATS_VERBOSE` and
    :setting:`DEPTH_PRIORITY` settings.

    .. reqmeta:: depth_reset

    depth_reset
    -----------

    .. versionadded:: 2.18.0

    :attr:`~scrapy.Request.meta` key that, set to ``True``, gives a request
    depth 0 instead of the depth of its source response plus 1, e.g. to keep
    :setting:`DEPTH_LIMIT` from applying across a domain change.
    """

    crawler: Crawler

    def __init__(  # pylint: disable=super-init-not-called
        self,
        maxdepth: int,
        stats: StatsCollector,
        verbose_stats: bool = False,
        prio: int = 1,
    ):
        self.maxdepth = maxdepth
        self.stats = stats
        self.verbose_stats = verbose_stats
        self.prio = prio
        self._ignored_logged = False

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        settings = crawler.settings
        maxdepth = settings.getint("DEPTH_LIMIT")
        verbose = settings.getbool("DEPTH_STATS_VERBOSE")
        prio = settings.getint("DEPTH_PRIORITY")
        o = cls(maxdepth, crawler.stats, verbose, prio)
        o.crawler = crawler
        return o

    @_warn_spider_arg
    def process_spider_output(
        self,
        response: Response | None,
        result: Iterable[Any],
        spider: Spider | None = None,
    ) -> Iterable[Any]:
        if response is not None:
            self._init_depth(response)
        yield from super().process_spider_output(response, result)

    @_warn_spider_arg
    async def process_spider_output_async(
        self,
        response: Response | None,
        result: AsyncIterator[Any],
        spider: Spider | None = None,
    ) -> AsyncIterator[Any]:
        if response is not None:
            self._init_depth(response)
        async for o in super().process_spider_output_async(response, result):
            yield o

    def _init_depth(self, response: Response) -> None:
        # base case (depth=0)
        if "depth" not in response.meta:
            response.meta["depth"] = 0
            if self.verbose_stats:
                self.stats.inc_value("request_depth_count/0")

    def get_processed_request(
        self, request: Request, response: Response | None
    ) -> Request | None:
        # Consumed here so that it cannot reach response.meta and, from there,
        # spread to further requests through a meta copy.
        depth_reset = request.meta.pop("depth_reset", False)
        if response is None:
            # start requests
            return request
        depth = 0 if depth_reset else response.meta["depth"] + 1
        request.meta["depth"] = depth
        if self.prio:
            request.priority -= depth * self.prio
        if self.maxdepth and depth > self.maxdepth:
            if not self._ignored_logged:
                logger.debug(
                    f"Ignoring link (depth > {self.maxdepth}): {request.url}"
                    " - no more ignored links will be shown",
                    extra={"spider": self.crawler.spider},
                )
                self._ignored_logged = True
            self.stats.inc_value("depth/request_ignored_count")
            return None
        if self.verbose_stats:
            self.stats.inc_value(f"request_depth_count/{depth}")
        self.stats.max_value("request_depth_max", depth)
        return request



import logging
import sys

import scrapy
from scrapy.crawler import AsyncCrawlerProcess
from scrapy.utils.reactorless import ReactorImportHook

logger = logging.getLogger(__name__)


def hook_count() -> int:
    return sum(1 for finder in sys.meta_path if isinstance(finder, ReactorImportHook))


class NoRequestsSpider(scrapy.Spider):
    name = "no_request"

    async def start(self):
        self.logger.info(f"Hooks during run: {hook_count()}")
        return
        yield


for _ in range(2):
    process = AsyncCrawlerProcess(settings={"TWISTED_REACTOR_ENABLED": False})
    process.crawl(NoRequestsSpider)
    process.start()

logger.info(f"Hooks after runs: {hook_count()}")



from __future__ import annotations

import logging
import warnings
from typing import cast
from unittest import mock

import OpenSSL._util as pyOpenSSLutil
import OpenSSL.SSL
import pytest
from OpenSSL import crypto

from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.utils.ssl import (
    _log_ssl_conn_debug_info,
    ffi_buf_to_string,
    get_openssl_version,
    get_temp_key_info,
    x509name_to_string,
)


def test_ffi_buf_to_string() -> None:
    buf = pyOpenSSLutil.ffi.new("char[]", b"some text")
    with pytest.warns(
        ScrapyDeprecationWarning, match=r"ffi_buf_to_string\(\) is deprecated"
    ):
        assert ffi_buf_to_string(buf) == "some text"


def test_x509name_to_string() -> None:
    with warnings.catch_warnings():
        # X509 itself is deprecated in pyOpenSSL, but it is the only way to build
        # the X509Name that the deprecated function under test takes.
        warnings.simplefilter("ignore", DeprecationWarning)
        subject = crypto.X509().get_subject()
    subject.C = "IE"
    subject.O = "Scrapy"
    subject.CN = "localhost"
    with pytest.warns(
        ScrapyDeprecationWarning, match=r"x509name_to_string\(\) is deprecated"
    ):
        assert x509name_to_string(subject) == "/C=IE/O=Scrapy/CN=localhost"


def test_get_temp_key_info() -> None:
    with pytest.warns(
        ScrapyDeprecationWarning, match=r"get_temp_key_info\(\) is deprecated"
    ):
        assert get_temp_key_info(object()) is None


def test_get_openssl_version() -> None:
    assert "OpenSSL" in get_openssl_version()


def test_log_ssl_conn_debug_info_no_certificate(
    caplog: pytest.LogCaptureFixture,
) -> None:
    connection = mock.MagicMock()
    connection.get_protocol_version_name.return_value = "TLSv1.3"
    connection.get_cipher_name.return_value = "TLS_AES_256_GCM_SHA384"
    connection.get_peer_certificate.return_value = None
    with caplog.at_level(logging.DEBUG, logger="scrapy.utils.ssl"):
        _log_ssl_conn_debug_info(
            "example.com", cast("OpenSSL.SSL.Connection", connection)
        )
    assert "SSL connection to example.com using protocol TLSv1.3" in caplog.text
    assert "certificate" not in caplog.text



from __future__ import annotations

import subprocess
import sys
from typing import TYPE_CHECKING, Any

import pytest

from scrapy.utils.test import get_testenv

if TYPE_CHECKING:
    from pathlib import Path

    from pexpect.popen_spawn import PopenSpawn


def call(*args: str, **popen_kwargs: Any) -> int:
    args = (sys.executable, "-m", "scrapy.cmdline", *args)
    return subprocess.call(
        args,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        env=get_testenv(),
        **popen_kwargs,
    )


def proc(
    *args: str, python_args: tuple[str, ...] = (), **popen_kwargs: Any
) -> tuple[int, str, str]:
    args = (sys.executable, *python_args, "-m", "scrapy.cmdline", *args)
    try:
        p = subprocess.run(
            args,
            check=False,
            capture_output=True,
            encoding="utf-8",
            timeout=15,
            env=get_testenv(),
            **popen_kwargs,
        )
    except subprocess.TimeoutExpired:
        pytest.fail("Command took too much time to complete")

    return p.returncode, p.stdout, p.stderr


def stop_spawn(p: PopenSpawn[str]) -> None:
    p.sendeof()
    p.wait()  # type: ignore[no-untyped-call]
    # PopenSpawn leaves the subprocess pipes open, which triggers
    # ResourceWarning at an arbitrary garbage collection point.
    for pipe in (p.proc.stdin, p.proc.stdout):
        if pipe:
            pipe.close()


def write_recording_editor(editor: Path) -> None:
    """Create an executable editor script that writes the path it is asked to
    open (its last argument) into the file given as its first argument."""
    editor.write_text('#!/bin/sh\nprintf "%s" "$2" > "$1"\n', encoding="utf-8")
    editor.chmod(0o755)


def write_recording_browser(browser: Path, recorded: Path) -> None:
    """Create an executable browser script that writes the URL it is asked to
    open into *recorded*.

    ``webbrowser`` only passes the URL to the command from the ``BROWSER``
    environment variable, hence the hardcoded output path.
    """
    browser.write_text(
        f'#!/bin/sh\nprintf "%s" "$1" > "{recorded}"\n', encoding="utf-8"
    )
    browser.chmod(0o755)



from __future__ import annotations

from scrapy.http import Response
from tests.utils.bases.http_response import TestResponseBase


class TestResponse(TestResponseBase):
    response_class = Response



import time
from typing import Any

import pytest

from scrapy import Request
from scrapy.core.downloader import Downloader, Slot
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.utils.spider import DefaultSpider
from scrapy.utils.test import get_crawler
from tests.mockserver.http import MockServer
from tests.spiders import MetaSpider
from tests.utils.decorators import coroutine_test, inline_callbacks_test


class DownloaderSlotsSettingsTestSpider(MetaSpider):
    name = "downloader_slots"

    custom_settings = {
        "DOWNLOAD_DELAY": 1,
        "RANDOMIZE_DOWNLOAD_DELAY": False,
        "DOWNLOAD_SLOTS": {
            "quotes.toscrape.com": {
                "concurrency": 1,
                "delay": 2,
                "randomize_delay": False,
                "throttle": False,
            },
            "books.toscrape.com": {"delay": 3, "randomize_delay": False},
        },
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        assert self.mockserver
        self.default_slot = self.mockserver.host
        self.times: dict[str, list[float]] = {}

    async def start(self):
        slots = [*self.custom_settings.get("DOWNLOAD_SLOTS", {}), None]
        for slot in slots:
            url = self.mockserver.url(f"/?downloader_slot={slot}")
            self.times[slot or self.default_slot] = []
            yield Request(url, callback=self.parse, meta={"download_slot": slot})

    def parse(self, response):
        slot = response.meta.get("download_slot", self.default_slot)
        self.times[slot].append(time.time())
        url = self.mockserver.url(f"/?downloader_slot={slot}&req=2")
        yield Request(url, callback=self.not_parse, meta={"download_slot": slot})

    def not_parse(self, response):
        slot = response.meta.get("download_slot", self.default_slot)
        self.times[slot].append(time.time())


class NoDelayDownloaderSlotsSpider(DownloaderSlotsSettingsTestSpider):
    custom_settings = {
        "DOWNLOAD_SLOTS": {
            slot: {}
            for slot in DownloaderSlotsSettingsTestSpider.custom_settings[
                "DOWNLOAD_SLOTS"
            ]
        },
    }


@inline_callbacks_test
def test_delay(mockserver: MockServer):
    crawler = get_crawler(DownloaderSlotsSettingsTestSpider)
    yield crawler.crawl(mockserver=mockserver)
    assert isinstance(crawler.spider, DownloaderSlotsSettingsTestSpider)
    slots = crawler.engine.downloader.slots
    times = crawler.spider.times
    # Downloading and processing a response add a roughly constant amount
    # of time on top of the configured delay, so the margin is absolute.
    # It stays well below the 1 second that separates the delays being
    # compared, so a slot using the delay of another one still fails.
    tolerance = 0.75

    for slot, (first, second) in times.items():
        assert abs((second - first) - slots[slot].delay) < tolerance


@coroutine_test
async def test_params():
    params = {
        "concurrency": 1,
        "delay": 2,
        "randomize_delay": False,
    }
    settings = {
        "DOWNLOAD_SLOTS": {
            "example.com": params,
        },
    }
    crawler = get_crawler(DefaultSpider, settings_dict=settings)
    crawler.spider = crawler._create_spider()
    downloader = Downloader(crawler)
    request = Request("https://example.com")
    _, actual = downloader._get_slot(request)
    downloader.close()
    expected = Slot(**params)
    for param in params:
        assert getattr(expected, param) == getattr(actual, param), (
            f"Slot.{param}: {getattr(expected, param)!r} != {getattr(actual, param)!r}"
        )


@coroutine_test
async def test_get_slot_deprecated_spider_arg():
    crawler = get_crawler(DefaultSpider)
    crawler.spider = crawler._create_spider()
    downloader = Downloader(crawler)
    request = Request("https://example.com")

    with pytest.warns(
        ScrapyDeprecationWarning,
        match=r"Passing a 'spider' argument to Downloader\._get_slot\(\) is deprecated",
    ):
        key1, slot1 = downloader._get_slot(request, spider=crawler.spider)
    key2, slot2 = downloader._get_slot(request)
    downloader.close()

    assert key1 == key2
    assert slot1 == slot2


@pytest.mark.parametrize(
    "priority_queue_class",
    [
        "scrapy.pqueues.ScrapyPriorityQueue",
        "scrapy.pqueues.DownloaderAwarePriorityQueue",
    ],
)
@coroutine_test
async def test_none_slot_with_priority_queue(
    mockserver: MockServer, priority_queue_class: str
) -> None:
    """Test specific cases for None slot handling with different priority queues."""
    crawler = get_crawler(
        NoDelayDownloaderSlotsSpider,
        settings_dict={"SCHEDULER_PRIORITY_QUEUE": priority_queue_class},
    )
    await crawler.crawl_async(mockserver=mockserver)
    assert isinstance(crawler.spider, NoDelayDownloaderSlotsSpider)

    assert hasattr(crawler.spider, "times")
    assert None not in crawler.spider.times
    assert crawler.spider.default_slot in crawler.spider.times
    assert len(crawler.spider.times[crawler.spider.default_slot]) == 2

    stats = crawler.stats
    assert stats.get_value("spider_exceptions", 0) == 0
    assert stats.get_value("downloader/exception_count", 0) == 0
